/*
** Module   : DynamicTablesHelper.java
** Abstract : Helper class for dynamic temporary tables.
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SVL 20130624 Created initial version.
** 002 SVL 20130822 Continued implementation.
** 003 SVL 20131115 Implemented tableExists, some refactoring for dynamic buffers registration
**                  and clean up.
** 004 CA  20131013 Added no-op deleted() method, required by the changes in Finalizable interface.
** 005 OM  20140117 Added support for runtime indexed access to fields from buffer.
** 006 SVL 20140123 Implemented get4GLTableName.
** 007 SVL 20140124 Refactored DMO class cleanup.
** 008 OM  20140123 Added legacy table name in createDynamicDMO. Also the fields ar normalized.
** 009 OM  20140128 Fixed historical annotation so the table/fields can be find by P2OLookup.
**                  Removed ConcurrentModificationException in cleanup.
** 010 ECF 20140226 Renamed generateDMOAst to generateSchemaAst. Added index ASTs to the schema
**                  AST produced by this method. Replaced Apache commons logging with J2SE
**                  logging.
** 011 ECF 20140404 Reimplemented dynamic conversion to use ConversionPool.
** 012 OM  20140414 Added case-sensitive field support for indexed character columns.
*/

package com.goldencode.p2j.persist;

import antlr.*;
import com.goldencode.ast.*;
import com.goldencode.compile.*;
import com.goldencode.p2j.convert.*;
import com.goldencode.p2j.pattern.*;
import com.goldencode.p2j.persist.dialect.P2JDialect;
import com.goldencode.p2j.schema.*;
import com.goldencode.p2j.security.ContextLocal;
import com.goldencode.p2j.uast.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager;
import com.goldencode.p2j.xml.*;
import org.hibernate.cfg.*;
import org.hibernate.dialect.Dialect;
import org.hibernate.internal.*;
import org.hibernate.mapping.*;
import org.w3c.dom.Document;
import java.io.*;
import java.util.*;
import java.util.logging.*;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.atomic.*;

/**
 * Helper class for dynamic temporary tables.
 */
public class DynamicTablesHelper
{
   /** Base package for generated DMOs. */
   public final static String DMO_BASE_PACKAGE = "com.goldencode.p2j.persist.dynamic";

   /** Category name for DMO implementation classes, used in conversion rules. */
   private final static String DMO_IMPL_CATEGORY = "impl";

   /** Category name for DMO interfaces, used in conversion rules. */
   private final static String DMO_IFACE_CATEGORY = "iface";

   /** Logger. */
   private final static Logger LOG = LogHelper.getLogger(DynamicTablesHelper.class);

   /** Singleton instance. */
   private static volatile DynamicTablesHelper instance;

   /** Param type to Progress type. */
   private final static Map<ParmType, Integer> propTypeToProgress = new HashMap<>();

   /** Containers for context-specific data. */
   private final static ContextLocal<Context> context = new ContextLocal<Context>()
   {
      public Context initialValue()
      {
         return new Context();
      }
   };

   /**
    * Reference to an instance of {@link InMemoryClassLoader} that presents in the chain
    * of system class loaders. If it doesn't present, the value is <code>null</code>.
    */
   private static InMemoryClassLoader systemInMemoryClassLoader;

   /** Counter used for generation of unique names for dynamic tables. */
   private final AtomicLong tableCounter = new AtomicLong(1);

   /** Shared Hibernate <code>Configuration</code> object. */
   private final Configuration cfg =
      DatabaseManager.getConfigBySchema(DatabaseManager.TEMP_TABLE_SCHEMA);


   static
   {
      propTypeToProgress.put(ParmType.CHAR, ProgressParserTokenTypes.FIELD_CHAR);
      propTypeToProgress.put(ParmType.INT, ProgressParserTokenTypes.FIELD_INT);
      propTypeToProgress.put(ParmType.INT64, ProgressParserTokenTypes.FIELD_INT64);
      propTypeToProgress.put(ParmType.DEC, ProgressParserTokenTypes.FIELD_DEC);
      propTypeToProgress.put(ParmType.LOG, ProgressParserTokenTypes.FIELD_LOGICAL);
      propTypeToProgress.put(ParmType.DATE, ProgressParserTokenTypes.FIELD_DATE);
      propTypeToProgress.put(ParmType.DT, ProgressParserTokenTypes.FIELD_DATETIME);
      propTypeToProgress.put(ParmType.DTTZ, ProgressParserTokenTypes.FIELD_DATETIME_TZ);
      propTypeToProgress.put(ParmType.HANDLE, ProgressParserTokenTypes.FIELD_HANDLE);
      propTypeToProgress.put(ParmType.RECID, ProgressParserTokenTypes.FIELD_RECID);
      propTypeToProgress.put(ParmType.ROWID, ProgressParserTokenTypes.FIELD_ROWID);
      propTypeToProgress.put(ParmType.RAW, ProgressParserTokenTypes.FIELD_RAW);

      systemInMemoryClassLoader = InMemoryClassLoader.getSystemInMemoryClassLoader();
      if (systemInMemoryClassLoader == null)
               throw new IllegalStateException("It is expected to have InMemoryClassLoader " +
                                               "in the chain of system class loaders.");
   }

   /**
    * Get singleton instance of the helper.
    *
    * @return  singleton instance of the helper.
    */
   public static DynamicTablesHelper getInstance()
   {
      if (instance == null)
      {
         synchronized (DynamicTablesHelper.class)
         {
            if (instance == null)
            {
               instance = new DynamicTablesHelper();
               PatternEngine.setDebugLevel(ConversionDriver.MSG_NONE);
            }
         }
      }

      return instance;
   }

   /**
    * Get Hibernate field name by 4GL field name for a dynamic DMO.
    *
    * @param  dmoIface
    *         DMO interface that corresponds the target dynamic DMO class.
    * @param  name4gl
    *         4GL field name.
    *
    * @return Hibernate field name for the given DMO class and 4GL field name or <code>null</code>
    *         if no corresponding name was found.
    */
   public static String getHibernateFieldName(Class dmoIface, String name4gl)
   {
      Map<String, String> map = context.get().fields4GLtoHibernate.get(dmoIface);
      if (map == null)
         return null;

      return map.get(normalizeName(name4gl));
   }

   /**
    * Get Hibernate field name by field creation id for a dynamic DMO.
    *
    * @param  dmoIface
    *         DMO interface that corresponds the target dynamic DMO class.
    * @param  fieldId
    *         4GL field id.
    *
    * @return Hibernate field name for the given DMO class and 4GL field name or <code>null</code>
    *         if no corresponding name was found.
    */
   public static String getHibernateFieldName(Class dmoIface, int fieldId)
   {
      // the algorithm here is based on the way the hibernate field names are generated at
      // "conversion" time. See (construction of clsFieldsXXX mappings in createDynamicDMO())
      return String.format("field%d", fieldId);
   }

   /**
    * Get 4GL field name by Hibernate field name for a dynamic DMO.
    *
    * @param  dmoIface
    *         DMO interface that corresponds the target dynamic DMO class.
    * @param  fieldNameHibernate
    *         Hibernate field name.
    *
    * @return 4gl field name for the given DMO class and Hibernate field name or <code>null</code>
    *         if no corresponding name was found.
    */
   public static String get4GLFieldName(Class dmoIface, String fieldNameHibernate)
   {
      Map<String, String> map = context.get().fieldsHibernateTo4GL.get(dmoIface);
      if (map == null)
         return null;

      return map.get(fieldNameHibernate);
   }

   /**
    * Normalize 4GL field or index name: drop trailing spaces and convert to lower case. Equality
    * tests for field or index names should be done in their normalized form.
    *
    * @param  name
    *         Field or index name to normalize.
    *
    * @return Normalized form of the field or index name.
    */
   public static String normalizeName(String name)
   {
      return TextOps.rightTrim(name.toLowerCase()).getValue();
   }

   /**
    * Remove trailing spaces from the string.
    *
    * @param  str
    *         Target string.
    *
    * @return the target string without trailing spaces.
    */
   public static String rightTrim(String str)
   {
      return TextOps.rightTrim(str).getValue();
   }

   /**
    * Get table name and database name from the string which represents full table name
    *
    * @param fullTableName
    *        The string which represents full table name. Can have two formats:
    *        "databaseName.tableName" or "tableName".
    *
    * @return String array with two elements. The first element is table name. The second element
    *         is database name or <code>null</code> if the database name wasn't specified in the
    *         full table name.
    */
   public static String[] parseTableName(String fullTableName)
   {
      String tableName;
      String databaseName = null;

      int pos = rightTrim(fullTableName).indexOf('.');
      if (pos < 0)
      {
         tableName = rightTrim(fullTableName);
      }
      else
      {
         // TODO database name should be case-insensitive, but for now it is considered to be
         // case-sensitive
         databaseName = rightTrim(fullTableName.substring(0, pos));
         tableName = rightTrim(fullTableName.substring(pos + 1));
      }

      return new String[] {tableName, databaseName};
   }

   /**
    * Allows to determine if the buffer is for a dynamic table.
    *
    * @param  buffer
    *         Buffer to check.
    *
    * @return <code>true</code> if buffer is for a dynamic table.
    */
   public static boolean isDynamicTableBuffer(RecordBuffer buffer)
   {
      TempTable tempTable = buffer.getParentTable();
      return tempTable != null && tempTable._dynamic();
   }

   /**
    * Get 4GL name of the table which corresponds the given DMO interface class.
    *
    * @param  dmoIface
    *         DMO interface class.
    *
    * @return see above.
    */
   public static String get4GLTableName(Class dmoIface)
   {
      TempTableBuilder ttb = getTableInfo(dmoIface);
      if (ttb != null)
      {
         // dynamic table
         return ttb.name().getValue();
      }

      String schema = DatabaseManager.getSchemaByInterface(dmoIface);
      Class<?> dmoClass = DMOIndex.getImplementingClass(schema, dmoIface);

      LegacyTable legacyTable = dmoClass.getAnnotation(LegacyTable.class);
      return legacyTable.name();
   }

   /**
    * Returns {@link TempTableBuilder} object that contains information about 4GL representation
    * of the table or <code>null</code> if there is no corresponding dynamic table for the
    * specified DMO interface.
    *
    * @param  dmoIface
    *         DMO interface that corresponds the target dynamic table.
    *
    * @return See above.
    */
   static TempTableBuilder getTableInfo(Class dmoIface)
   {
      return context.get().tables.get(dmoIface);
   }

   /**
    * Determines if the specified table exists.
    *
    * @param  persistence
    *         Persistence which corresponds the target database.
    * @param  tableName4GL
    *         4GL table name.
    *
    * @return <code>true</code> if the specified table exists.
    */
   static boolean tableExists(Persistence persistence, String tableName4GL)
   {
      String normalizedTableName = normalizeName(tableName4GL);
      Database database = persistence.getDatabase();

      if (database.equals(DatabaseManager.TEMP_TABLE_DB))
      {
          return BufferManager.get().getStaticTempTableDepth(normalizedTableName) != -1;
      }
      else
      {
         String schema = DatabaseManager.getSchema(database);
         return TableMapper.getDMOClass(schema, normalizedTableName) != null;
      }
   }

   /**
    * Search for an existing table with the specified name.
    *
    * @param  fullTableName
    *         Table name in format "databaseName.tableName" or "tableName". In the latter case
    *         table search is performed in all permanent databases and then the temporary
    *         database. Dynamic tables are ignored.
    *
    * @return persistence which corresponds the target database or <code>null</code> if there is
    *         no such table.
    */
   static Persistence findExistingTable(String fullTableName)
   {
      String[] arr = parseTableName(fullTableName);
      String srcTableName = arr[0];
      String srcDatabaseName = arr[1];

      if (srcDatabaseName != null)
      {
         // database is explicitly specified
         if (!ConnectionManager.connected(srcDatabaseName).booleanValue())
            return null;

         String physicalName = ConnectionManager.pdbName(srcDatabaseName).getValue();
         Persistence persistence = PersistenceFactory.getInstance(physicalName);

         if (tableExists(persistence, srcTableName))
            return persistence;
         else
            return null;
      }
      else
      {
         // search through all databases, check permanent databases first
         long numDbs = ConnectionManager.numDBs().getValue();
         for (int i = 1; i <= numDbs; i++)
         {
            String databaseName = ConnectionManager.ldbName(i).getValue();
            if (!ConnectionManager.connected(databaseName).booleanValue())
            {
               continue;
            }

            String physicalName = ConnectionManager.pdbName(databaseName).getValue();
            Persistence persistence = PersistenceFactory.getInstance(physicalName);

            if (!tableExists(persistence, srcTableName))
            {
               continue;
            }

            return persistence;
         }

         // check temporary database
         Persistence persistence =
                  PersistenceFactory.getInstance(DatabaseManager.TEMP_TABLE_DB);

         if (tableExists(persistence, srcTableName))
            return persistence;
         else
            return null;
      }
   }

   /**
    * Get original 4GL table name. Despite the fact that we specify 4GL table name as an input
    * parameter, this parameter can have different letter case or may have trailing spaces (this
    * form is allowed when specifying table name in 4GL code).
    *
    * @param  persistence
    *         Persistence which corresponds the target database.
    * @param  tableName4GL
    *         4GL table name (case-insensitive, may have trailing spaces).
    *
    * @return original 4GL table name.
    */
   static String get4GLTableName(Persistence persistence, String tableName4GL)
   {
      // TODO implement when table names in the original letter case will be preserved
      return normalizeName(tableName4GL);
   }
   
   /**
    * Generate DMO interface, implementation class and mapping document and register the new DMO
    * type.
    *
    * @param   builder
    *          Temporary table builder containing definitions of the target table.
    * @param   legacyTableName
    *          The legacy name of the table.
    *
    * @return  <code>true</code> on success.
    */
   public boolean createDynamicDMO(TempTableBuilder builder, String legacyTableName)
   {
      long tableId = tableCounter.getAndIncrement();
      String hibernateTableName = String.format("dtt%d", tableId);
      
      String ifaceName = String.format("DynamicRecord%d", tableId);
      String implClassName = String.format("DynamicRecord%dImpl", tableId);
      Map<String, String> clsFields4GLtoHibernate = new HashMap<>();
      Map<String, String> clsFieldsHibernateTo4GL = new HashMap<>();
      
      long fieldId = 1;
      for (P2JField field : builder.getFields())
      {
         String fieldName4GL = normalizeName(field.getName());
         String fieldNameHibernate = String.format("field%d", fieldId++);
         clsFields4GLtoHibernate.put(fieldName4GL, fieldNameHibernate);
         clsFieldsHibernateTo4GL.put(fieldNameHibernate, fieldName4GL);
      }
      
      Aast root = instance.generateSchemaAst(builder,
                                             ifaceName,
                                             hibernateTableName,
                                             normalizeName(legacyTableName),
                                             clsFields4GLtoHibernate);
      
      try
      {
         ConversionPool.Results results;
         results = ConversionPool.runTask(ConversionProfile.P2O, root);
         
         root = (Aast) results.getStoredObject(null);
         root.putAnnotation("permanent", false);
         root.putAnnotation("package", DMO_BASE_PACKAGE);
         
         // generate .hbm.xml
         results = ConversionPool.runTask(ConversionProfile.HIBERNATE, root);
         XmlAst hbmXml = (XmlAst) results.getStoredObject(ifaceName);
         
         // generate interface and implementation Java ASTs
         results = ConversionPool.runTask(ConversionProfile.JAVA_DMO, root);
         Aast impl = (Aast) results.getStoredObject(DMO_IMPL_CATEGORY, ifaceName);
         Aast iface = (Aast) results.getStoredObject(DMO_IFACE_CATEGORY, ifaceName);
         
         // generate Java code from Java ASTs
         results = ConversionPool.runTask(ConversionProfile.BREW, impl, iface);
         
         ByteArrayOutputStream implCodeStream =
            (ByteArrayOutputStream) results.getStoredObject(implClassName);
         ByteArrayOutputStream ifaceCodeStream =
            (ByteArrayOutputStream) results.getStoredObject(ifaceName);
         
         String fullIfaceName = String.format("%s.%s.%s",
                                              DMO_BASE_PACKAGE,
                                              DatabaseManager.TEMP_TABLE_SCHEMA,
                                              ifaceName);
         String fullImplClassName = String.format("%s.%s.impl.%s",
                                                  DMO_BASE_PACKAGE,
                                                  DatabaseManager.TEMP_TABLE_SCHEMA,
                                                  implClassName);
         StringBuilder sb = new StringBuilder();
         Map<String, Class<?>> classes = InMemoryCompiler.compile(
                        new String[] { fullIfaceName, fullImplClassName },
                        new String[] { ifaceCodeStream.toString(), implCodeStream.toString() },
                        sb,
                        true);
         
         if (classes == null)
         {
            LOG.log(Level.SEVERE, sb.toString());
            ErrorManager.displayError("Cannot compile Hibernate entities.");
            return false;
         }
         
         Context local = context.get();
         
         Class<?> dmoIface = classes.get(fullIfaceName);
         Class<?> dmoClass = classes.get(fullImplClassName);
         local.loadedClasses.put(dmoIface, classes.values());
         
         MutableSessionFactoryImpl sessionFactory = (MutableSessionFactoryImpl)
            DatabaseManager.getSessionFactory(DatabaseManager.TEMP_TABLE_DB);
         PersistentClass persistentClass;
         
         // collect the list of character fields that will take part in indexes
         Set<String> indexedColumns = new HashSet<>();
         Set<String> caseSensitiveCols = new HashSet<>();
         for (P2JIndex index4GL : builder.getIndexes())
         {
            Iterator<P2JIndexComponent> iter = index4GL.components();
            while (iter.hasNext())
            {
               P2JIndexComponent comp4GL = iter.next();
               // add character fields that are part of an index
               if (comp4GL.isCharType())
               {
                  final String dtFieldName = clsFields4GLtoHibernate.get(comp4GL.getName());
                  indexedColumns.add(dtFieldName);
                  if (!comp4GL.isIgnoreCase())
                  {
                     caseSensitiveCols.add(dtFieldName);
                  }
               }
            }
         }
         
         synchronized (cfg)
         {
            // SVL: as far as I've checked Configuration object is not accessed by Hibernate
            // outside synchronized blocks in this class.  DBUtils.makeTypeArray (which also
            // uses Configuration object), accesses Configuration.typeResolver which remains
            // intact in the synchronized blocks.
            
            Document document = XmlAst.astToDom(hbmXml);
            P2JDialect dialect =
                  (P2JDialect) DatabaseManager.getDialect(DatabaseManager.TEMP_TABLE_DB);
            ORMHandler.walkChildren(indexedColumns, caseSensitiveCols, document, dialect);
            cfg.addDocument(document);
            cfg.buildMappings();
            persistentClass = cfg.getClassMapping(dmoClass.getName());
            
            sessionFactory.add(persistentClass, cfg);
         }
         
         // collect index names generated by conversion, map them by legacy name
         Map<String, String> idxp2j = new HashMap<>();
         LegacyIndexes lindexes = dmoClass.getAnnotation(LegacyIndexes.class);
         if (lindexes != null)
         {
            for (LegacyIndex lidx : lindexes.value())
            {
               idxp2j.put(lidx.name().toLowerCase(), lidx.converted());
            }
         }
         
         // translate 4GL indexes to Hibernate indexes
         List<P2JIndex> indexesHibernate = new ArrayList<>();
         for (P2JIndex index4GL : builder.getIndexes())
         {
            String legacyIndexName = index4GL.getName().toLowerCase();
            String indexNameHibernate = String.format("idx__%s", idxp2j.get(legacyIndexName));
            P2JIndex indexHibernate = new P2JIndex(hibernateTableName,
                                                   indexNameHibernate,
                                                   index4GL.isUnique(),
                                                   index4GL.isPrimary(),
                                                   index4GL.isWord());
            Iterator<P2JIndexComponent> iter = index4GL.components();
            while (iter.hasNext())
            {
               P2JIndexComponent comp4GL = iter.next();
               String fieldName = clsFields4GLtoHibernate.get(normalizeName(comp4GL.getName()));
               indexHibernate.addComponent(fieldName,
                                           comp4GL.isDescending(),
                                           comp4GL.isCharType(),
                                           comp4GL.isIgnoreCase(),
                                           comp4GL.isAbbreviated());
            }
            
            indexesHibernate.add(indexHibernate);
         }
         
         DMOIndex.registerDynamicTable(dmoIface, dmoClass, indexesHibernate);
         DatabaseManager.registerDynamicTable(dmoIface, persistentClass);
         
         DatabaseManager.mapComputedCharacterColumns(
               DatabaseManager.TEMP_TABLE_SCHEMA,
               DatabaseManager.TEMP_TABLE_DB,
               cfg);
         
         local.fields4GLtoHibernate.put(dmoIface, clsFields4GLtoHibernate);
         local.fieldsHibernateTo4GL.put(dmoIface, clsFieldsHibernateTo4GL);
         local.tables.put(dmoIface, builder);
         
         builder.setDmoIface(dmoIface);
         builder.setPersistentClass(persistentClass);
      }
      catch (Exception e)
      {
         LOG.log(Level.SEVERE, e.getMessage(), e);
         ErrorManager.displayError("Error while generating dynamic table: " + e.getMessage());
         return false;
      }
      
      return true;
   }
   
   /**
    * Drop table, all buffers and all information about the specified dynamic DMO class.
    *
    * @param dmoIface
    *        DMO interface class.
    */
   public void dropTable(Class dmoIface)
   {
      Context local = context.get();
      Set<TemporaryBuffer> buffers = local.dynamicTableBuffers.get(dmoIface);
      int lastIndex = buffers.size() - 1;
      int i = 0;
      for (TemporaryBuffer buffer : buffers)
      {
         buffer.release();
         boolean dropTable = (i == lastIndex);

         // drop table during cleanup of the last iterated buffer for the DMO interface
         buffer.closeMultiplexScope(dropTable);
         buffer.finished();
         i++;
      }

      dropDynamicDMO(dmoIface, true);
   }

   /**
    * Remove information about the specified dynamic DMO class from Hibernate Configuration,
    * session factory and P2J facilities.
    *
    * @param dmoIface
    *        DMO interface which corresponds the target dynamic DMO class.
    * @param cleanContextLocal
    *        If <code>true</code> then context-local information in P2J facilities will be cleaned
    *        (if may want not do that if the session is closing because all context-local
    *        information will be dropped anyway).
    */
   public void dropDynamicDMO(Class dmoIface, boolean cleanContextLocal)
   {
      Context local = context.get();
      TempTableBuilder table = local.tables.get(dmoIface);
      PersistentClass persistentClass = table.getPersistentClass();

      MutableSessionFactoryImpl sessionFactory =
         (MutableSessionFactoryImpl) DatabaseManager.getSessionFactory(DatabaseManager.TEMP_TABLE_DB);
      sessionFactory.remove(persistentClass);

      synchronized (cfg)
      {
         cfg.remove(persistentClass);
      }

      for (Class cls : local.loadedClasses.get(dmoIface))
      {
         systemInMemoryClassLoader.deregister(cls.getName());
      }

      if (cleanContextLocal)
      {
         Class dmoClass = persistentClass.getMappedClass();
         local.tables.remove(dmoIface);
         local.fields4GLtoHibernate.remove(dmoIface);
         local.fieldsHibernateTo4GL.remove(dmoIface);
         local.loadedClasses.remove(dmoIface);
         local.dynamicTableBuffers.remove(dmoIface);
         DMOIndex.deregisterDynamicTable(dmoIface, dmoClass);
         DatabaseManager.deregisterDynamicTable(dmoIface, persistentClass);
      }
   }

   /**
    * Register the dynamic buffer for a dynamic table. That will ensure buffer clean up if the
    * parent table is deleted.
    *
    * @param  buffer
    *         Dynamic buffer for a dynamic table to be registered.
    */
   public void registerBufferForDynamicTable(TemporaryBuffer buffer)
   {
      Context local = context.get();
      Class dmoIface = buffer.getDMOInterface();
      Set<TemporaryBuffer> buffers = local.dynamicTableBuffers.get(dmoIface);
      if (buffers == null)
      {
         buffers = new HashSet<TemporaryBuffer>();
         local.dynamicTableBuffers.put(dmoIface, buffers);
      }
      buffers.add(buffer);
   }

   /**
    * Generate AST that represents the target table and has the same structure as .p2o files from
    * the conversion stage.
    *
    * @param  builder
    *         Temporary table builder containing definitions of the target table.
    * @param  ifaceName
    *         The name of the target DMO interface.
    * @param  tableName
    *         The name of the target SQL table.
    * @param  legacyTableName
    *         The lecagy (Progress) name of the table.
    * @param  fields4GLtoHibernate
    *         Map 4GL field names to Hibernate names.
    *
    * @return the root node of the generated AST.
    */
   private Aast generateSchemaAst(TempTableBuilder builder,
                                  String ifaceName,
                                  String tableName,
                                  String legacyTableName,
                                  Map<String, String> fields4GLtoHibernate)
   {
      Aast root = new ProgressAst(new CommonToken(ProgressParserTokenTypes.DATABASE,
                                                  DatabaseManager.TEMP_TABLE_SCHEMA));
      
      ProgressAst cls = new ProgressAst(new CommonToken(ProgressParserTokenTypes.TABLE,
                                                        tableName));
      cls.putAnnotation("class-name", ifaceName);
      cls.putAnnotation("table-name", tableName);
      cls.putAnnotation(P2OLookup.HISTORICAL, legacyTableName);
      root.addChild(cls);
      
      // add field nodes to table
      Map<String, Aast> fieldAsts = new HashMap<>();
      for (P2JField field : builder.getFields())
      {
         String fieldName4GL = normalizeName(field.getName());
         String fieldNameHibernate = fields4GLtoHibernate.get(fieldName4GL);
         ParmType type = field.getType();
         int progressType = propTypeToProgress.get(type);
         ProgressAst property = new ProgressAst(new CommonToken(progressType, fieldNameHibernate));
         fieldAsts.put(fieldName4GL, property);
         property.putAnnotation("format", field.getFormat());
         property.putAnnotation("label", field.getLabel());
         property.putAnnotation(P2OLookup.HISTORICAL, fieldName4GL);
         if (field.isCaseSensitive())
         {
            property.putAnnotation("case-sensitive", field.isCaseSensitive());
         }
         
         long extent = field.getExtent();
         if (extent > 1)
            property.putAnnotation("extent", extent);
         
         cls.addChild(property);
         
         // add initial value node to field
         BaseDataType initial = field.getInitial();
         if (initial != null)
         {
            ProgressAst initialNode = new ProgressAst(
                                      new CommonToken(ProgressParserTokenTypes.KW_INIT, ""));
            property.addChild(initialNode);
            
            int valType;
            if (initial.isUnknown())
               valType = ProgressParserTokenTypes.UNKNOWN_VAL;
            else
            {
               switch (type)
               {
                  case CHAR:
                     valType = ProgressParserTokenTypes.STRING;
                     break;
                  case INT:
                     valType = ProgressParserTokenTypes.NUM_LITERAL;
                     break;
                  case DEC:
                     valType = ProgressParserTokenTypes.DEC_LITERAL;
                     break;
                  case LOG:
                     valType = initial.equals(new logical(true)) ?
                           ProgressParserTokenTypes.BOOL_TRUE :
                           ProgressParserTokenTypes.BOOL_FALSE;
                     break;
                  default:
                     throw new IllegalStateException();
               }
            }
            
            String valStr;
            switch (type)
            {
               case CHAR:
                  valStr = "\"" + initial.toStringMessage() + "\"";
                  break;
               default:
                  valStr = initial.toStringMessage();
            }
            
            ProgressAst initialVal = new ProgressAst(new CommonToken(valType, valStr));
            initialNode.addChild(initialVal);
         }
      }
      
      // add index nodes to table
      for (P2JIndex index : builder.getIndexes())
      {
         String indexName = normalizeName(index.getName());
         ProgressAst indexAst = new ProgressAst(new CommonToken(ProgressParserTokenTypes.INDEX,
                                                                indexName));
         cls.addChild(indexAst);
         
         if (index.isPrimary())
         {
            indexAst.putAnnotation(P2OLookup.PRIMARY, true);
         }
         
         if (index.isUnique())
         {
            indexAst.putAnnotation(P2OLookup.UNIQUE, true);
         }
         
         if (index.isWord())
         {
            indexAst.putAnnotation(P2OLookup.WORD, true);
         }
         
         // add index fields to index
         Iterator<P2JIndexComponent> compIter = index.components();
         while (compIter.hasNext())
         {
            P2JIndexComponent comp = compIter.next();
            String compName = normalizeName(comp.getName());
            ProgressAst compAst = new ProgressAst(
               new CommonToken(ProgressParserTokenTypes.INDEX_FIELD, compName));
            indexAst.addChild(compAst);
            
            if (comp.isDescending())
            {
               compAst.putAnnotation(P2OLookup.DESCEND, true);
            }
            
            // TODO: what about ABBREVIATED and CASE-SENSITIVE options (not handled as part of
            // TRPL schema fixups today)?
         }
      }
      
      assignIds(root, 1);
      
      // now assign each index field's refid to the corresponding field's ID 
      Aast indexAst = cls.getImmediateChild(ProgressParserTokenTypes.INDEX, null);
      while (indexAst != null)
      {
         Aast compAst = indexAst.getImmediateChild(ProgressParserTokenTypes.INDEX_FIELD, null);
         while (compAst != null)
         {
            String fieldName4GL = compAst.getText();
            Aast fieldAst = fieldAsts.get(fieldName4GL);
            compAst.putAnnotation("refid", fieldAst.getId());
            compAst = indexAst.getImmediateChild(ProgressParserTokenTypes.INDEX_FIELD, compAst);
         }
         
         indexAst = cls.getImmediateChild(ProgressParserTokenTypes.INDEX, indexAst);
      }
      
      return root;
   }

   /**
    * Traverse an AST and assign some unique ids to all of its nodes.
    *
    * @param  root
    *         The root node of the AST.
    * @param  startId
    *         First available id for assignment.
    *
    * @return new first available id for assignment.
    */
   private long assignIds(Aast root, long startId)
   {
      root.setId(startId++);

      int childCount = root.getNumImmediateChildren();
      if (childCount > 0)
      {
         for (int i = 0; i < childCount; i++)
         {
            Aast child = root.getChildAt(i);
            startId = assignIds(child, startId);
         }
      }
      return startId;
   }

   /**
    * Container for context-local information about dynamic tables and buffers.
    */
   private static class Context
   implements Finalizable
   {
      /**
       * Mapping of <i>normalized</i> 4GL field names to Hibernate names ("field1", "field2",
       * ...). The outermost keys are corresponding DMO interface classes.
       */
      private final Map<Class, Map<String, String>> fields4GLtoHibernate = new HashMap<>();

      /**
       * Mapping of Hibernate field names ("field1", "field2", ...) to 4GL field names. The
       * outermost keys are corresponding DMO interface classes.
       */
      private final Map<Class, Map<String, String>> fieldsHibernateTo4GL = new HashMap<>();

      /**
       * Mapping of DMO interface classes to {@link TempTableBuilder} objects that contain
       * information about 4GL representation of the corresponding tables
       */
      private final Map<Class, TempTableBuilder> tables = new HashMap<>();

      /**
       * Sets containing all dynamic record buffers created in this context keyed by DMO interface
       * classes.
       */
      private final Map<Class, Set<TemporaryBuffer>> dynamicTableBuffers = new HashMap<>();

      /**
       * Mapping of DMO interfaces to classes loaded by InMemoryClassLoader for this DMO
       * (DMO interface and implementation classes, Buf class, Composite classes).
       */
      private final Map<Class, Collection<Class<?>>> loadedClasses = new HashMap<>();

      /**
       * Constructor. Registers this instance of {@link Context} as a global finalizable.
       */
      private Context()
      {
         TransactionManager.registerFinalizable(this, true);
      }

      /**
       * Clean up local context - drop tables and DMO definitions.
       */
      public void finished()
      {
         // use a copy of the collection in order to avoid the iterator exception
         // java.util.ConcurrentModificationException caused by indirect removing the currently
         // iterated dmoIface in dropTable() / dropDynamicDMO()
         Set<Class> keys = tables.keySet();
         Class[] classes = new Class[keys.size()];
         classes = keys.toArray(classes); // should return the same array object
         for (Class dmoIface : classes)
         {
            instance.dropTable(dmoIface);
         }
      }

      /** No-op. */
      public void deleted()
      {
      }

      /** No-op. */
      public void iterate()
      {
      }

      /** No-op. */
      public void retry()
      {
      }
   }
}
