/*
** Module   : HQLPreprocessor.java
** Abstract : Rewrites HQL where clauses under certain conditions
**
** Copyright (c) 2004-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20060317   @25230 Created initial version. Identifies certain
**                           patterns in a where clause string, which
**                           indicate unsupported HQL "shortcut" syntax,
**                           and rewrites the where clause to be compliant
**                           with supported, HQL syntax.
** 002 ECF 20060713   @28049 Expanded purpose of this class to modify HQL
**                           to trim trailing whitespace from all text
**                           properties. A SQL RTRIM function is inserted
**                           around each qualified reference to a text
**                           property.
** 003 ECF 20060814   @28689 Added support for ternary clauses. These are
**                           implemented using CASE WHEN... clauses.
** 004 ECF 20060901   @29152 Changed exception processing. We now throw
**                           IllegalArgumentException with a verbose
**                           message from preprocessTextNodes() upon
**                           encountering an unrecognized alias.
** 005 ECF 20060906   @29257 Workaround for Hibernate HQL parser defect.
**                           The 'THEN' and 'ELSE' component expressions
**                           of a 'CASE WHEN' clause must be enclosed in
**                           parentheses to prevent the parser from
**                           blowing up.
** 006 ECF 20061004   @30153 Fix for trailing trim (rtrim) processing.
**                           Functions which return text must be enclosed
**                           within the rtrim() function as well.
** 007 ECF 20061020   @30588 Added support for string concatenation
**                           operator (||). Added to the set of supported,
**                           binary operators.
** 008 ECF 20061030   @30792 Commented code which rewrites unary logical
**                           expressions. This code would rewrite all such
**                           expressions as binary, equality comparisons
**                           with the logical constant 'true', in order to
**                           work around a defect in the Hibernate HQL
**                           parser. That defect has now been patched,
**                           rendering the workaround obsolete with our
**                           modified version of Hibernate 3.0.5. However,
**                           moving to a standard Hibernate release which
**                           does not have this patch will require that
**                           this code be reactivated.
** 009 ECF 20061119   @31324 Added collection of restriction criteria
**                           property names. These are needed when
**                           processing ChangeBroker notifications about
**                           DMO state changes to determine whether a
**                           preselected result set should be invalidated
**                           by AdaptiveQuery.
** 010 ECF 20061121   @31344 Fixed NPE. Added safety code to new method
**                           getRestrictionProperties().
** 011 ECF 20070328   @32645 Added caching of HQLPreprocessor instances.
**                           Made constructor private and added factory
**                           methods to retrieve cached or new instances.
** 012 ECF 20070502   @33370 Added limited support for subselect phrases.
**                           Only those features required to support the
**                           server-side implementation of converted,
**                           nested CAN-FIND statements are implemented.
**                           Also removed need to pass RecordBuffer
**                           instances to factory method.
** 013 ECF 20070723   @34653 Fixed generation of ANSI join clauses. In
**                           cases where multiple different indices in a
**                           composite inner class were being referenced,
**                           we were only creating a single join for that
**                           composite.
** 014 ECF 20071216   @36492 Added injection of error handling functions
**                           around certain expressions. This allows
**                           proper error handling of converted CAN-FIND
**                           statements which are nested within a query's
**                           where clause.
** 015 ECF 20080306   @37477 Added significant, dialect-driven capability.
**                           Added support for computed columns. In cases
**                           where a database dialect does not support
**                           embedding functions within indices but uses
**                           computed instead, we replace certain function
**                           signatures within where clauses with
**                           references to those computed columns. This
**                           presumably enables the database's query
**                           planner to choose matching indices more
**                           easily. Added "manual" user defined function
**                           overloading for cases where the backing
**                           dialect does not support natural overloading.
** 016 ECF 20080328   @37724 Fixed mainWalk(). An attempt to check the
**                           type of a standalone alias (i.e., no property
**                           child) was causing an NPE in DataTypeHelper.
** 017 ECF 20080330   @37738 Modified to support method change in
**                           DatabaseManager.
** 018 ECF 20080508   @38485 Added support for function overloading in
**                           dirty databases. Functions are now registered
**                           by Database object rather than database name.
** 019 ECF 20080606   @38606 Added new factory method. Accepts Database
**                           P2JDialect args rather than Persistence. This
**                           is necessary for dirty database use of this
**                           class, where there is no Persistence instance
**                           associated with a dirty database. Changed
**                           CacheKey to use Database instead of
**                           Persistence for the same reason.
** 020 ECF 20080817   @39555 Map restriction properties to DMO entity name
**                           instead of DMO implementation class.
** 021 ECF 20080930   @40078 Added substitution parameter inlining. Query
**                           substitution parameters which represent an
**                           operand in subexpressions which perform range
**                           checks (i.e., >, >=, <, <=, some forms of
**                           LIKE) may be inlined into the where clause.
**                           With certain database dialects, this may
**                           allow better query plans to be chosen, at the
**                           expense of reparsing the query.
** 022 ECF 20081015   @40092 Fixed regression caused by #021 (@40078).
**                           Ensure substitution parameters are properly
**                           preprocessed to avoid trailing whitespace in
**                           strings and out-of-range date values.
** 023 ECF 20081015   @40094 Fixed inlineSubstitutionParameter(). Use a
**                           dialect-specific date format to avoid query
**                           failure with BC dates.
** 024 ECF 20081031   @40307 Fixed parameter inlining and caching. Certain
**                           HQLPreprocessor instances were being cached
**                           without taking inlining status into account.
**                           Also, inlineSubstitutionParameter() must
**                           resolve FieldReferences, but not other types
**                           of Resolvables.
** 025 SVL 20090118   @41170 Comparisons with the unknown are handled
**                           properly.
** 026 SVL 20090121   @41188 <= ? is translated to "true" and < ? is
**                           translated to "is not null".
** 027 SVL 20090122   @41198 Do not look into cache if we have a query
**                           with unknown parameter(s).
** 028 ECF 20090310   @41501 Fixed performance problem with is [not] null
**                           refactoring. In cases where such a test is
**                           made against a non-nullable column/property,
**                           the expression can be optimized to avoid any
**                           record reads. For example, 'table.id is null'
**                           must always be false and 'table.id is not
**                           null' must always be true. With some database
**                           implementations, this avoids expensive query
**                           plan choices.
** 029 SVL 20090415   @41797 Added roll up of the expressions containing
**                           booleans (during HQL tree processing).
** 030 GES 20090421   @41830 Matched utility class package change.
** 031 SVL 20090428   @42028 Conjunction of index(alias_compositeXXXX)=Y
**                           conditions is added to the top level of the AST
**                           (instead of keeping this conditions linked to the
**                           original conditions and therefore spreading
**                           associations between the original conditions on
**                           them too).
** 032 GES 20090429   @42050 Match package and class name changes.
** 033 GES 20090518   @42390 Import change.
** 034 ECF 20090601   @42584 Added support for Progress isolation leak quirk.
**                           The execution of some queries (the current
**                           understanding is any where clause which checks a
**                           database field for inequality with any operand)
**                           will trigger the premature publication of all
**                           uncommitted changes in related DMO types to all
**                           sessions. The publication is handled by the dirty
**                           database manager, but the identification of the
**                           trigger conditions is managed here.
** 035 ECF 20090617   @42764 Fixed function overloading. Overloading was not
**                           being performed in cases where no substitution
**                           parameters were in use.
** 036 ECF 20090716   @43218 Added support for lazy initialization of record
**                           buffers. resolveBuffer() now calls buffer's
**                           initialize() method, to ensure buffers referenced
**                           in subselects are initialized.
** 037 SVL 20090810   @43577 "then ? else ?" statement is replaced with "then
**                           cast(? as <datatype>) else cast(? as <datatype>)"
**                           for H2 dialect.
** 038 SVL 20110622          Convert \\ to \ for the H2 dialect. Call
**                           postprocessStringLiteral for newly inlined string
**                           literals.
** 039 SVL 20110814   @43700 Improved unknown handling for "? <comparison
**                           operator> {const|subst}", "? <comparison
**                           operator> ?" and standalone unknowns.
** 040 SVL 20120331          Upgraded to Hibernate 4. Make explicit cast in ternary also when
**                           we have a single substitution parameter. Added handling for DATETIME
**                           and DATETIMETZ types.
** 041 OM  20130602          Added support for named parameters for datetime-tz datatype.
**                           Added preprocessing to inject session attributes into hql.
** 042 CA  20130822          The DATE-FORMAT attribute must be accessed via SessionUtils.
** 043 SVL 20140210          Use HQLExpression instead of HQL string.
** 044 VMN 20140328          Added support for custom denormalization of fields with extent.
** 045 ECF 20140403          Integrated Aast API changes; replaced Apache commons logging with
**                           J2SE logging.
** 046 OM  20140410          Fixed support case-sensitive fields in indexes.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import java.util.logging.*;
import org.hibernate.type.Type;
import com.goldencode.ast.*;
import com.goldencode.util.*;
import com.goldencode.p2j.persist.hql.*;
import com.goldencode.p2j.persist.dialect.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.LogHelper;

/**
 * A preprocessor for where clauses written in (extended) HQL.  It performs
 * the following functions:
 * <ul>
 *   <li>Handles references to text properties in a dialect-specific manner.
 *       In cases where functions are usable within index definitions,
 *       encloses each qualified reference to a text value within a SQL
 *       RTRIM function, so that trailing whitespace (specifically space,
 *       tab, line feed, and carriage return characters) is trimmed by the
 *       database server at query execution time.  This is necessary to
 *       emulate Progress' treatment of character fields in where clauses.
 *       Qualifying references may be DMO properties which store text values
 *       or server-side functions which return text data.  In cases where
 *       functions cannot be embedded within index definitions, replaces text
 *       property references with computed column references.
 *   <li>Rewrites where clauses written with an enhanced form of HQL syntax
 *       on the fly, in order to make it compliant with the HQL syntax
 *       supported by Hibernate.
 *   <li>Rewrites unary logical expressions as binary comparisons with
 *       <code>true</code> in order to work around a defect in the Hibernate
 *       HQL parser which disallows unary logical conditions.
 *   <li>Identifies and preserves sets of DMO properties used as restriction
 *       criteria in the preprocessed where clause.  The names of these
 *       properties are stored as sets, indexed in a map by the associated
 *       DMO implementation class(es).
 *   <li>Injects error handling functions around expressions which need to be
 *       protected.  Expressions which meet the following criteria must be
 *       enclosed in the <code>checkError()</code> function.  The first
 *       parameter to this function is <code>initError(false)</code>, and the
 *       second parameter is the expression requiring error handling.  The
 *       criteria are:
 *       <ul>
 *         <li>include a server-side PL function;
 *         <li>evaluate to a boolean result;
 *         <li>reside within the where clause of a subselect phrase.
 *       </ul>
 *       <p>
 *       The call to <code>initError(false)</code> initializes the error
 *       handler for the current <code>checkError()</code> scope.  The
 *       execution of the expression may or may not report an error to the
 *       error handler.  The <code>checkError()</code> function checks the
 *       error handler's current scope for an error and handles it as
 *       necessary.
 *   <li>Handles "manual" overloading of user defined functions. Multiple user
 *       defined functions can be registered for the same base functions name.
 *       At preprocessing time, the parameter signature determines which
 *       overloaded function will be used, and the overloaded function name is
 *       replaced in HQL with a specific backing function.
 *   <li>In certain cases, inlines query substitution parameters directly into
 *       the where clause string.  Only those parameters which participate in
 *       range matches (e.g., &gt;, &gt;=, &lt;, &lt;=, and some forms of
 *       LIKE) are inlined, and then only if the database dialect supports it,
 *       and the caller has requested it.  Instances of this class which have
 *       had parameters inlined are not cached.  The purpose of inlining is to
 *       allow the backing database to choose a better query plan when it
 *       prepares the query, since it has more information about range matches
 *       than if this information were provided later, at query execution
 *       time.
 * </ul>
 * <p>
 * Furthermore, each instance of this class provides the necessary contextual
 * information required by the enclosing query to properly embed the
 * rewritten where clause into an overall HQL statement.
 * <p>
 * For example, the P2J environment allows the following syntax:
 * <pre>
 *    from SomeDMOImpl as alias
 *    where alias.propertyABC[5] = ?
 * </pre>
 * as shorthand for something like:
 * <pre>
 *    from SomeDMOImpl as alias
 *    join alias.composite32 as alias_composite32
 *    where alias_composite32.propertyABC = ? and index(alias_composite32) = 5
 * </pre>
 * <p>
 * The former syntax is not directly supported by Hibernate and so the where
 * clause <code>alias.propertyABC[5] = ?</code> is expanded to the latter
 * form internally, within the P2J persistence framework.
 * <p>
 * This rewriting is only necessary in the particular situation where a DMO
 * class contains a list of composite elements, and a where clause references
 * a property of such a composite element.  The above example corresponds to
 * a Hibernate mapping of:
 * <pre>
 *    ...
 *    [class name="SomeDMOImpl" ...]
 *       ...
 *       [list name="composite32" ]
 *          ...
 *          [composite-element class="SomeDMOImpl$Composite32"]
 *             [property name="propertyABC" ... /]
 *             ...
 *          [/composite-element]
 *       [/list]
 *       ...
 *    [/class]
 *    ...
 * </pre>
 * <p>
 * Allowing the shorthand syntax hides the complexity of the DMO
 * implementation underlying the Hibernate mapping, and prevents the public
 * API of the persistence runtime from becoming more complicated.  The latter
 * point is made because, in addition to the expansion of the where clause,
 * one or more ANSI-style joins will be added to the overall HQL statement,
 * and the order of query substitution parameters may be changed.  The
 * following service methods are provided to enable enclosing queries to
 * embed the preprocessed where clause in the HQL statement, and to organize
 * substitution parameters in the proper order:
 * <ul>
 *   <li>{@link #getHQL()}
 *   <li>{@link #getParameterIndices()}
 *   <li>{@link #ansiJoins()}
 * </ul>
 * <p>
 * A where clause is preprocessed during construction of an instance of this
 * class.  Parsing errors are logged, but are otherwise ignored.  Ultimately,
 * Hibernate will report an HQL problem if the input where clause is invalid.
 */
public final class HQLPreprocessor
implements HQLParserTokenTypes
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(HQLPreprocessor.class);
   
   /** A do-nothing instance for empty where clauses */
   private static final HQLPreprocessor NOP_INSTANCE;
   
   /** Cache of prepared instances, indexed by Persistence and where clause */
   private static final Map<CacheKey, HQLPreprocessor> cache = new HashMap<>();
   
   /** Map of function keys to overloaded function names */
   private static final Map<FunctionKey, String> overloadedFunctions = new HashMap<>();
   
   static
   {
      try
      {
         NOP_INSTANCE = new HQLPreprocessor(null, null, null, null, null, null, false, false);
      }
      catch (PersistenceException exc)
      {
         throw new RuntimeException(exc);
      }
   }
   
   /** DMO class (only for a single buffer where clause without aliases) */
   private final Class<?> dmoClass;
   
   /** Database associated with this object */
   private final Database database;
   
   /** HQL subexpressions to perform ANSI joins of associated lists */
   private LinkedHashSet<String> ansiJoins = null;
   
   /** HQL-compliant where clause (possibly rewritten) */
   private HQLExpression hql = null;
   
   /** Whether substitution parameter inlining is permitted/was performed */
   private boolean inline = false;
   
   /** Mapping of index positions for substitution parameters after rewrite */
   private ParameterIndices paramIndices = null;
   
   /** Map of properties used by this where clause, keyed by DMO entity */
   private Map<String, Set<String>> restrictionProperties = null;
   
   /** List of entities which publish uncommitted changes prematurely */
   private List<String> earlyPublishEntities = null;
   
   /**
    * Factory method which accepts an unprocessed where clause, the database
    * instance and the database dialect associated with the enclosing query,
    * the DMO implementation class, and the substitution parameters, if any,
    * for the query.  This method may return a shared, immutable instance from
    * a cache if such an object was created previously for the given where
    * clause and parameter types.
    *
    * @param   where
    *          An unprocessed where clause which will be scanned, parsed, and
    *          possibly rewritten.  Should not include the leading
    *          <code>where</code> keyword.
    * @param   database
    *          Database object associated with the client query.
    * @param   dialect
    *          Database dialect in use.
    * @param   dmoClass
    *          DMO implementation class associated with the query, if the
    *          where clause does not qualify its property references with
    *          aliases.
    * @param   parameters
    *          Query substitution parameters for the where clause.
    * @param   inline
    *          <code>true</code> to permit inlining of substitution parameters
    *          involved in range matches (&gt;, &gt;=, &lt;, &lt;=, like);
    *          <code>false</code> to disallow such inlining.  Inlining will
    *          embed such parameters into the HQL string directly.
    *
    * @return  An instance of this class.  At the time this object is
    *          returned, all preprocessing work is complete.
    *
    * @throws  PersistenceException
    *          if there is any error rewriting the where clause.
    */
   public static HQLPreprocessor get(String where,
                                     Database database,
                                     P2JDialect dialect,
                                     Class<?> dmoClass,
                                     Object[] parameters,
                                     boolean inline)
   throws PersistenceException
   {
      if (where == null)
      {
         return NOP_INSTANCE;
      }
      
      HQLPreprocessor preproc = null;
      Type[] paramTypes = null;
      boolean overload = !dialect.supportsFunctionOverloading();
      boolean unknowns = false;
      Object[] paramArray = parameters;
      if (parameters != null)
      {
         paramTypes = DBUtils.makeTypeArray(parameters);
         inline = inline && dialect.isQueryRangeParameterInlined();
         
         // Resolve all parameters and check whether we have any unknown
         // value(s).
         int len = parameters.length;
         for (int i = 0; i < len; i++)
         {
            Object parm = parameters[i];
            if (parm instanceof FieldReference)
            {
               if (paramArray == parameters)
               {
                  paramArray = new Object[len];
                  System.arraycopy(parameters, 0, paramArray, 0, len);
               }
               parm = ((FieldReference) parm).resolve();
               paramArray[i] = parm;
            }
            
            if (parm instanceof BaseDataType && ((BaseDataType) parm).isUnknown())
            {
               unknowns = true;
            }
         }
      }
      
      CacheKey key = new CacheKey(database, dmoClass, where, paramTypes, inline);
      if (!unknowns)
      {
         synchronized (cache)
         {
            preproc = cache.get(key);
         }
      }
      
      // This synchronization approach may result in duplicate instances of
      // the preprocessor being created in separate threads for the same
      // where clause.  However, this would occur only in relatively rare
      // race condition cases.  Better that than having every thread block on
      // the global monitor while any preprocessor does its grunt work.
      if (preproc == null)
      {
         preproc = new HQLPreprocessor(database,
                                       where,
                                       dmoClass,
                                       paramArray,
                                       paramTypes,
                                       dialect,
                                       overload,
                                       inline);
         
         if (!preproc.wasInlined())
         {
            // Only cache if no parameters were inlined, since inlining would
            // change the HQL string, breaking the hash key.  Also, it would
            // cause non-reusable preprocessors to bloat the cache.
            synchronized (cache)
            {
               cache.put(key, preproc);
            }
         }
      }
      
      return preproc;
   }
   
   /**
    * Factory method which accepts an unprocessed where clause, the record
    * buffer with which the enclosing query is associated (if the where clause
    * does not use aliases), and the query substitution parameters, if any,
    * for the query.  This method may return a shared, immutable instance from
    * a cache if such an object was created previously for the given where
    * clause and parameter types.
    *
    * @param   where
    *          An unprocessed where clause which will be scanned, parsed, and
    *          possibly rewritten.  Should not include the leading
    *          <code>where</code> keyword.
    * @param   buffer
    *          RecordBuffer associated with the query, if the where clause
    *          does not qualify its property references with aliases.
    * @param   parameters
    *          Query substitution parameters for the where clause.
    * @param   inline
    *          <code>true</code> to permit inlining of substitution parameters
    *          involved in range matches (&gt;, &gt;=, &lt;, &lt;=, like);
    *          <code>false</code> to disallow such inlining.  Inlining will
    *          embed such parameters into the HQL string directly.
    * 
    * @return  An instance of this class.  At the time this object is
    *          returned, all preprocessing work is complete.
    *
    * @throws  PersistenceException
    *          if there is any error rewriting the where clause.
    */
   static HQLPreprocessor get(String where,
                              RecordBuffer buffer,
                              Object[] parameters,
                              boolean inline)
   throws PersistenceException
   {
      Persistence persistence = buffer.getPersistence();
      Database database = persistence.getDatabase();
      P2JDialect dialect = (P2JDialect) persistence.getDialect();
      
      return get(where,
                 database,
                 dialect,
                 buffer.getDMOImplementationClass(),
                 parameters,
                 inline);
   }
   
   /**
    * Register a user-defined function to the specified backing method for the
    * given database.
    * 
    * @param   database
    *          Database for which user-defined function is registered.
    * @param   function
    *          Function name.
    * @param   method
    *          Method which will back the given function.
    */
   public static void registerFunction(Database database, String function, Method method)
   {
      Class<?>[] paramTypes = method.getParameterTypes();
      int len = paramTypes.length;
      HQLTypes[] signature = new HQLTypes[len];
      for (int i = 0; i < len; i++)
      {
         signature[i] = DataTypeHelper.getTypeClass(paramTypes[i]);
      }
      String name = method.getName();
      FunctionKey key = new FunctionKey(database, name, signature);
      
      if (overloadedFunctions.containsKey(key))
      {
         throw new IllegalArgumentException(
            "User defined function '" + key + "' already registered");
      }
      
      overloadedFunctions.put(key, function);
      
      if (LOG.isLoggable(Level.FINE))
      {
         LOG.log(Level.FINE, "Registering function '" + key + "' to '" + function + "'");
      }
   }
   
   /**
    * Constructor which accepts an unprocessed where clause.
    * 
    * @param   database
    *          Database within which the associated query will be run.
    * @param   where
    *          An unprocessed where clause which will be scanned, parsed, and
    *          possibly rewritten.  Should not include the leading
    *          <code>where</code> keyword.
    * @param   dmoClass
    *          DMO implementation class for where clauses which do not use
    *          aliases;  <code>null</code> if the where clause uses aliases.
    * @param   parameters
    *          Query substitution parameters for the where clause.
    * @param   paramTypes
    *          Hibernate query substitution parameter types.
    * @param   dialect
    *          Database dialect in use.
    * @param   overload
    *          <code>true</code> to manually overload user-defined functions.
    * @param   inline
    *          <code>true</code> to allow substitution parameters involved in
    *          range checks to be inlined;  <code>false</code> to disallow
    *          such inlining.  If inlining occurs, the preprocessor will not
    *          be cached.
    *
    * @throws  PersistenceException
    *          if there is any error rewriting the where clause.
    */
   private HQLPreprocessor(Database database,
                           String where,
                           Class<?> dmoClass,
                           Object[] parameters,
                           Type[] paramTypes,
                           P2JDialect dialect,
                           boolean overload,
                           boolean inline)
   throws PersistenceException
   {
      this.database = database;
      this.dmoClass = dmoClass;
      this.inline = inline;
      
      if (where != null)
      {
         preprocess(where, parameters, paramTypes, dialect, overload);
      }
      
      if (earlyPublishEntities == null)
      {
         earlyPublishEntities = Collections.emptyList();
      }
   }
   
   /**
    * Get the (possibly rewritten) where clause, which is compliant with HQL
    * syntax.
    *
    * @return  HQL where clause, with no leading <code>where</code> keyword.
    */
   HQLExpression getHQL()
   {
      return hql;
   }
   
   /**
    * Get the <code>ParameterIndices</code> object associated with this query.
    * This object maintains an array of zero-based indices into the array of
    * query substitution arguments for the where clause, as it exists
    * <i>after</i> preprocessing is complete.  This additional level of
    * indirection is necessary because the query substitution placeholders may
    * have been reordered within the HQL where clause expression during the
    * HQL preprocessing rewrite step, and some parameters may have been
    * inlined into the where clause.
    *
    * @return  An object which provides a mapping of indices into the query
    *          substitution parameter list.
    */
   ParameterIndices getParameterIndices()
   {
      return paramIndices;
   }
   
   /**
    * Get an iterator on the ANSI-style join subexpressions which must be
    * merged by the enclosing query into the overall HQL query statement, in
    * order to properly join associated composite element lists (if any).
    *
    * @return  An iterator on join subexpressions.  The iterator may be
    *          empty, but will not be <code>null</code>.
    */
   Iterator<String> ansiJoins()
   {
      return (ansiJoins != null
              ? ansiJoins.iterator()
              : new EmptyIterator<String>());
   }
   
   /**
    * Retrieve the list of DMO entity names which will trigger the premature
    * publication of uncommitted changes across sessions, once the query
    * containing this preprocessor's underlying where clause is executed.
    * <p>
    * This method is intended to support the emulation of a quirk/bug in
    * Progress whereby the execution of a query (in the generic sense of the
    * term) in a session after uncommitted changes, will trigger those changes
    * to be prematurely published to other sessions.
    * <p>
    * TODO:  this behavior may represent a version-specific bug in Progress
    * which should be disabled in P2J for other versions than 9.1C.
    * 
    * @return  The list of entities, if any, which will have their uncommitted
    *          changes published prematurely if the underlying where clause is
    *          executed in the current session.  The returned list may be
    *          empty, but it will never be <code>null</code>.
    */
   List<String> getEarlyPublishEntities()
   {
      return earlyPublishEntities;
   }
   
   /**
    * Get an unmodifiable map of DMO entity names to sets of the names of the
    * DMO properties used as restriction criteria in the preprocessed where
    * clause.
    *
    * @return  Map of DMO entity names to sets of property names, or
    *          <code>null</code> if the where clause contained no restriction
    *          properties.
    */
   Map<String, Set<String>> getRestrictionProperties()
   {
      return (restrictionProperties != null
              ? Collections.unmodifiableMap(restrictionProperties)
              : null);
   }
   
   /**
    * Indicate whether query substitution parameters participating in range
    * checks were inlined directly into the query string.  If called during
    * construction (before inlining actually occurs), this method will return
    * whether this object <i>can</i> be inlined.  If called after construction
    * completes, it will return whether this object actually <i>was</i>
    * inlined.  That is, even if inlining was permitted, it will not have
    * occurred if no suitable subexpressions were detected in the where clause
    * provided.
    * 
    * @return  <code>true</code> if inlining was performed (with the caveats
    *          noted above);  else <code>false</code>.
    */
   boolean wasInlined()
   {
      return inline;
   }
   
   /**
    * Resolve the record buffer associated with the given alias.  This method
    * delegates to the {@link BufferManager} to retrieve the buffer.
    *
    * @param   alias
    *          AST representing the alias (variable name) with which the
    *          buffer was defined in business logic.
    *
    * @return  Record buffer associated with the alias.
    *
    * @throws  IllegalArgumentException
    *          if <code>alias</code> does not match any buffer whose scope is
    *          currently open.
    */
   private RecordBuffer resolveBuffer(Aast alias)
   {
      String varName = alias.getText();
      RecordBuffer buffer = BufferManager.get().lookupBuffer(varName);
      if (buffer == null)
      {
         throw new IllegalArgumentException(
            "Unrecognized buffer at column " +
            alias.getColumn() +
            " in where clause:  '" +
            varName +
            "' [" +
            hql +
            "]");
      }
      
      buffer.initialize();
      
      return buffer;
   }
   
   /**
    * The main worker method which drives preprocessing.  Performs the
    * following steps:
    * <ol>
    *   <li>parse the input where clause;
    *   <li>restructure ASTs representing qualified names of text properties
    *       and functions to embed these names with an SQL function which
    *       trims trailing whitespace;
    *   <li>restructure ASTs representing unary logical expressions to
    *       instead be binary logical comparisons with <code>true</code>.
    *   <li>analyze the resulting AST to identify subtrees which require
    *       restructuring;
    *   <li>possibly inline certain query substitution parameters;
    *   <li>restructure the AST to account for the joins;
    *   <li>generate any necessary ANSI join subexpressions;
    *   <li>emit the rewritten result as a string and stores it for later
    *       access;
    *   <li>create the substitution parameter index mapping.
    * <ol>
    * <p>
    * If the analysis performed in step 3 determines that rewriting is not
    * necessary, the following three steps are skipped, and the original
    * where clause simply is stored.
    *
    * @param   where
    *          An unprocessed where clause which will be scanned, parsed, and
    *          possibly rewritten.  Should not include the leading
    *          <code>where</code> keyword.
    * @param   parameters
    *          Query substitution parameters for the where clause.
    * @param   paramTypes
    *          Hibernate query substitution parameter types.
    * @param   dialect
    *          Database dialect in use.
    * @param   overload
    *          <code>true</code> to manually overload user-defined functions.
    *
    * @throws  PersistenceException
    *          if there is any error rewriting the where clause.
    */
   private void preprocess(String where,
                           Object[] parameters,
                           Type[] paramTypes,
                           P2JDialect dialect,
                           boolean overload)
   throws PersistenceException
   {
      this.hql = new HQLExpression(where);
      HQLAst root = parse();
      if (root != null)
      {
         // Annotate query substitution parameter nodes with their data types.
         if (paramTypes != null)
         {
            annotateParameters(root, paramTypes);
         }
         
         // Walk the AST, fixing up certain nodes in place.
         root = mainWalk(root, parameters, paramTypes, dialect, overload);
         
         /* See header entry #008
         // Work around a defect in the Hibernate HQL parser which disallows
         // unary logical conditions.
         root = preprocessBooleanExpressions(root);
         */

         Iterable<HQLAst> denormalizedFields = getDenormalizedFields(root);
         if (denormalizedFields != null)
         {
            for (HQLAst property : denormalizedFields)
            {
               inlineDenormalizedField(property, parameters);
            }

            if (LOG.isLoggable(Level.FINEST))
            {
               String sep = System.getProperty("line.separator");
               LOG.log(Level.FINEST,
                       "[" + database + "] Denormalized HQL:  " + hql + sep + root.dumpTree());
            }
         }
         
         Map<HQLAst, String> targets = analyzeComposites(root);
         if (targets != null)
         {
            // Restructure the target ASTs.
            root = restructure(root, targets);
            
            // ANSI joins must be generated after restructure, since targets
            // are modified during restructuring to have appropriate composite names.
            this.ansiJoins = generateJoins(targets);
         }
         
         // Emit the preprocessed HQL expression text.
         this.hql = emit(root);

         if (LOG.isLoggable(Level.FINEST))
         {
            String sep = System.getProperty("line.separator");
            LOG.log(Level.FINEST,
                    "[" + database + "] Preprocessed HQL:  " + hql + sep + root.dumpTree());
         }
         
         this.paramIndices = createParameterIndices(root, parameters);
      }
   }
   
   /**
    * Scan and parse the input where clause and return it in the form of an
    * abstract syntax tree.  Rewriting is not performed at this stage.
    *
    * @return  AST representing the original where clause.
    * 
    * @throws  PersistenceException
    *          if there is any error parsing the HQL.
    */
   private HQLAst parse()
   throws PersistenceException
   {
      HQLAst root = null;
      
      try
      {
         StringReader reader = new StringReader(hql.toFinalExpression());
         HQLLexer lexer = new HQLLexer(reader);
         HQLParser parser = new HQLParser(lexer);
         
         parser.expression();
         root = (HQLAst) parser.getAST();
         root.fixups(null, null);
         
         if (LOG.isLoggable(Level.FINEST))
         {
            String sep = System.getProperty("line.separator");
            LOG.log(Level.FINEST,
                    "[" + database + "] Original HQL:  " + hql + sep + root.dumpTree());
         }
      }
      catch (Exception exc)
      {
         throw new PersistenceException("Error parsing HQL", exc);
      }
      
      return root;
   }
   
   /**
    * Walk the HQL AST, annotating query substitution parameter nodes with
    * their data types as they are visited.  The annotation name used for this
    * is <code>datatype</code>.
    * 
    * @param   root
    *          Root AST node at which we begin the walk.
    * @param   paramTypes
    *          Hibernate query substitution parameter types.
    */
   private void annotateParameters(HQLAst root, Type[] paramTypes)
   {
      int counter = 0;
      int len = paramTypes.length;
      
      Iterator<Aast> iter = root.iterator();
      while (iter.hasNext() && counter < len)
      {
         Aast ast = iter.next();
         if (ast.getType() == SUBST)
         {
            Type hibType = paramTypes[counter];
            HQLTypes hqlType = DataTypeHelper.getTypeParameter(hibType);
            ast.putAnnotation("datatype", hqlType.toString());
            counter++;
         }
      }
   }
   
   /**
    * Walk the AST depth-first from root to leaves and rewrite certain nodes.
    * The following modifications are made in place in the given AST:
    * <ul>
    *   <li>A node which represents the qualified name of a DMO text property
    *       is restructured such that a new node representing an SQL trimming
    *       function is inserted between the qualified property name and its
    *       existing parent.  A string node representing the characters to be
    *       trimmed is inserted as the property name's next sibling (i.e.,
    *       the second parameter to the trimming function).
    *   <li>The first child of a FROM node represents the unqualified DMO
    *       interface name provided by business logic as part of a subselect
    *       phrase.  This must be replaced with the unqualified name of the
    *       the DMO implementation class which implements this interface.
    *       The second child of the FROM node represents the record buffer
    *       alias which uniquely identifies the record buffer in scope.  This
    *       information is used to retrieve the buffer and determine the DMO
    *       implementation name to use.
    * </ul>
    *
    * @param   root
    *          Root node of the AST for the HQL expression.
    * @param   parameters
    *          Query substitution parameters for the where clause.
    * @param   paramTypes
    *          Hibernate query substitution parameter types.
    * @param   dialect
    *          Database dialect in use.
    * @param   overload
    *          <code>true</code> to manually overload user-defined functions.
    *
    * @throws  PersistenceException
    *          if any error occurred during rewriting the AST.
    * @throws  IllegalArgumentException
    *          if an alias encountered in the where clause is not mapped in
    *          the buffer manager.
    *
    * @return  The new root node of the AST for the HQL expression (it may
    *          change during processing).
    */
   private HQLAst mainWalk(HQLAst root,
                           Object[] parameters,
                           Type[] paramTypes,
                           P2JDialect dialect,
                           boolean overload)
   throws PersistenceException
   {
      Set<HQLAst> needTrim = null;
      Set<HQLAst> needErrorHandling = null;
      Set<HQLAst> unknowns = null;
      Set<HQLAst> needSimplification = null;
      Set<HQLAst> needExplicitCastInsideTernary = null;
      Set<HQLAst> needSingleBackslash = null;

      boolean doInline = inline;
      inline = false;
      HQLAst parent;
      
      // Walk the tree and identify all nodes requiring attention.
      Iterator<Aast> iter = root.iterator();
      while (iter.hasNext())
      {
         HQLAst next = (HQLAst) iter.next();
         
         switch (next.getType())
         {
            case FUNCTION:
               
               HQLTypes funcType = null;
               
               if (next.ancestor(-1, SUBSELECT))
               {
                  HQLAst hit = null;
                  funcType = DataTypeHelper.expressionType(next, dmoClass);
                  if (funcType == HQLTypes.BOOLEAN)
                  {
                     hit = next;
                  }
                  else
                  {
                     parent = (HQLAst) next.getParent();
                     if (DataTypeHelper.expressionType(parent, dmoClass) == HQLTypes.BOOLEAN)
                     {
                        hit = parent;
                     }
                  }
                  
                  if (hit != null)
                  {
                     if (needErrorHandling == null)
                     {
                        needErrorHandling = new HashSet<HQLAst>();
                     }
                     needErrorHandling.add(hit);
                  }
               }
               
               if (funcType == null)
               {
                  funcType = DataTypeHelper.expressionType(next, dmoClass);
               }
               
               // Identify whether this is a text node which needs to be
               // reparented with a trimming function.
               if (funcType == HQLTypes.TEXT && !"upper".equals(next.getText()))
               {
                  if (needTrim == null)
                  {
                     needTrim = new HashSet<>();
                  }
                  needTrim.add(next);
               }
               
               // there is a bug in progress when calling pl/java functions with no arguments
               // it will display "To many parameters - expected 0"
               // however, we can "inline" the value of the static function at this moment by
               // replacing the FUNCTION node with an integer with actual value 
               // this is faster as no supplementary pl/java call will be done on sql-server side
               if (doInline && next.getNumImmediateChildren() == 0)
               {
                  if ("getMtime".equals(next.getText()))
                  {
                     next.setType(NUM_LITERAL);
                     next.setText(datetime.now().getTime() + "");
                     next.putAnnotation("inlined", Boolean.TRUE);
                     inline = true;
                  }
                  
                  if ("getTimezone".equals(next.getText()))
                  {
                     next.setType(NUM_LITERAL);
                     next.setText(date.getDefaultOffset() + "");
                     next.putAnnotation("inlined", Boolean.TRUE);
                     inline = true;
                  }
               }
               
               // toDate(string) / toDatetime(string) / toDatetimeTz(string) functions must be 
               // used in the 2nd form (with format) because the SESSION:DATE-FORMAT and 
               // WINDOWING-YEAR are not accessible sql-server-side
               if (next.getNumImmediateChildren() == 1    && 
                   next.getChildAt(0).getType() == STRING &&
                   ("toDate".equals(next.getText())     ||
                    "toDatetime".equals(next.getText()) ||
                    "toDatetimeTz".equals(next.getText())))
               {
                  HQLAst formatNode = new HQLAst();
                  formatNode.setType(STRING);
                  formatNode.setText('\'' + SessionUtils.getDateOrder().toStringMessage() + '\'');
                  next.graftAt(formatNode, 1); // as 2nd parameter
                  
                  HQLAst windowingNode = new HQLAst();
                  windowingNode.setType(NUM_LITERAL);
                  windowingNode.setText(date.getWindowingYear() + "");
                  next.graftAt(windowingNode, 2); // as 3rd parameter
               }
               
               // If manual function overloading is required, do so at this time.
               if (overload)
               {
                  manuallyOverload(next);
               }
               
               break;
               
            case ALIAS:
               
               HQLAst child0 = (HQLAst) next.getChildAt(0);
               
               // Identify whether this is a text node which needs to be
               // reparented with a trimming function or replaced with a
               // computed column reference.
               if (child0 != null &&   // omit standalone alias
                   DataTypeHelper.expressionType(next, dmoClass) == HQLTypes.TEXT)
               {
                  if (needTrim == null)
                  {
                     needTrim = new HashSet<HQLAst>();
                  }
                  needTrim.add(next);
               }
               
               // Determine whether this where clause represents a trigger for
               // the Progress quirk which leaks a snapshot of all uncommitted
               // changes to other sessions.  Record the affected entity in
               // the list of entities who will have their uncommitted changes
               // so published.
               if (child0 != null && child0.getType() == PROPERTY)
               {
                  parent = (HQLAst) next.getParent();
                  while (parent != null && !isBinaryOperator(parent))
                  {
                     parent = (HQLAst) parent.getParent();
                  }
                  
                  if (parent != null)
                  {
                     switch (parent.getType())
                     {
                        case NOT_EQ:
                           RecordBuffer buffer = resolveBuffer(next);
                           String entity = buffer.getEntityName().intern();
                           if (earlyPublishEntities == null)
                           {
                              earlyPublishEntities = new ArrayList<String>();
                           }
                           if (!earlyPublishEntities.contains(entity))
                           {
                              earlyPublishEntities.add(entity);
                           }
                           break;
                        
                        default:
                           break;
                     }
                  }
                  
               }
               
               continue;
               
            // Use the DMO alias to look up the DMO implementation name and
            // replace the text of the first child of a FROM clause.
            case FROM:
               
               HQLAst c0 = (HQLAst) next.getChildAt(0);
               HQLAst c1 = (HQLAst) next.getChildAt(1);
               if (c1 == null)
               {
                  // Create default alias by lowercasing the unqualified DMO name.
                  String name = c0.getText();
                  c1 = new HQLAst();
                  c1.setType(ALIAS);
                  c1.setText(StringHelper.changeCase(name, false));
               }
               RecordBuffer buffer = resolveBuffer(c1);
               c0.setText(buffer.getDMOImplementationName());
               
               continue;
               
            // Possibly inline certain query substitution parameters.
            case SUBST:
               
               parent = (HQLAst) next.getParent();
               if (parent != null)
               {
                  int index = ((Long) next.getAnnotation("index")).intValue();
                  Object param = parameters[index];
                  
                  if (param instanceof datetimetz)
                  {
                     // next.putAnnotation("positional-params", new Long(2));
                  }

                  // Check for unknown value.
                  if (param instanceof BaseDataType && ((BaseDataType) param).isUnknown())
                  {
                     if (unknowns == null)
                     {
                       unknowns = new HashSet<HQLAst>();
                     }

                     // if we have an unknown value as subscript, the extent
                     // field value is unknown (i.e. field[?] = ?), and we
                     // need to add it to the set instead of subscript
                     parent = (HQLAst) next.getParent();
                     if (parent != null && parent.getType() == LBRACKET)
                     {
                        parent = (HQLAst) parent.getParent();
                        if (parent != null && parent.getType() == PROPERTY)
                        {
                           HQLAst alias = (HQLAst) parent.getParent();
                           if (alias != null && alias.getType() == ALIAS)
                           {
                              next = alias;
                           }
                           else
                           {
                              next = parent;
                           }
                        }
                     }

                     // if unknown value has a wrapping function, value of
                     // this function is also unknown
                     while (true)
                     {
                        parent = (HQLAst) next.getParent();
                        if (parent != null && parent.getType() == FUNCTION)
                        {
                           next = parent;
                        }
                        else
                        {
                           break;
                        }
                     }

                     unknowns.add(next);

                     continue;
                  }

                  if (!doInline)
                  {
                     break;
                  }

                  switch (parent.getType())
                  {
                     case LIKE:
                     case GT:
                     case LT:
                     case GTE:
                     case LTE:
                     case OR:
                     case AND:
                        int tokenType = inlineSubstitutionParameter(next, parameters, dialect);
                        if (tokenType == BOOL_TRUE || tokenType == BOOL_FALSE)
                        {
                           if (needSimplification == null)
                           {
                              needSimplification = new HashSet<HQLAst>();
                           }
                           needSimplification.add(next);
                        }
                        else if (tokenType == STRING && dialect.usesSingleBackslash())
                        {
                           if (needSingleBackslash == null)
                           {
                              needSingleBackslash = new HashSet<HQLAst>();
                           }

                           needSingleBackslash.add(next);
                        }
                        break;
                        
                     default:
                        continue;
                  }
               }
               
               break;

            case TERNARY:
               if (dialect.requiresExplicitCastInsideTernary())
               {
                  if (needExplicitCastInsideTernary == null)
                  {
                     needExplicitCastInsideTernary = new HashSet<HQLAst>();
                  }
                  needExplicitCastInsideTernary.add(next);
               }
               break;

            case STRING:
               if (dialect.usesSingleBackslash())
               {
                  if (needSingleBackslash == null)
                  {
                     needSingleBackslash = new HashSet<HQLAst>();
                  }

                  needSingleBackslash.add(next);
               }
               break;

            default:
               continue;
         }
      }
      
      try
      {
         // Replace \\ with \ into string literals for some dialects
         // (e.g. H2 dialect). 
         if (needSingleBackslash != null)
         {
            for (HQLAst next : needSingleBackslash)
            {
               String replacement = next.getText().replace("\\\\", "\\");
               next.setText(replacement);
            }
         }

         // If any nodes were identified as needing the rtrim function,
         // reparent them with this function now or replace them with
         // computed column references.
         if (needTrim != null)
         {
            for (HQLAst next : needTrim)
            {
               parent = (HQLAst) next.getParent();
               boolean handled = false;
               
               // First try to use computed column references, if the dialect
               // calls for it.  Only valid for ALIAS nodes, not FUNCTION
               // nodes.
               if (next.getType() == ALIAS)
               {
                  // TODO:  handle unqualified properties.
                  handled = injectComputedColumn(next);
               }
               
               // If not handled with a computed column reference, inject an
               // rtrim() function call around the qualified property name.
               if (!handled)
               {
                  int idxPos = next.getIndexPos();
                  
                  HQLAst trimFunc = new HQLAst();
                  trimFunc.setType(FUNCTION);
                  trimFunc.setText("rtrim");
                  
                  HQLAst trimChars = new HQLAst();
                  trimChars.setType(STRING);
                  trimChars.setText("' \t\n\r'");
                  
                  next.remove();
                  parent.graftAt(trimFunc, idxPos);
                  trimFunc.graft(next);
                  trimFunc.graft(trimChars);
               }
            }
         }
         
         // If any nodes were identified as needing error handling, reparent
         // them with the checkError() function now.
         if (needErrorHandling != null)
         {
            for (HQLAst next : needErrorHandling)
            {
               int idxPos = next.getIndexPos();
               parent = (HQLAst) next.getParent();
               
               HQLAst checkErrFunc = new HQLAst();
               checkErrFunc.setType(FUNCTION);
               checkErrFunc.setText("checkError");
               
               HQLAst initErrFunc = new HQLAst();
               initErrFunc.setType(FUNCTION);
               initErrFunc.setText("initError");
               
               HQLAst falseParm = new HQLAst();
               falseParm.setType(BOOL_FALSE);
               falseParm.setText("false");
               initErrFunc.graft(falseParm);
               
               next.remove();
               parent.graftAt(checkErrFunc, idxPos);
               checkErrFunc.graft(initErrFunc);
               checkErrFunc.graft(next);
            }
         }

         // get rid of unknowns 
         if (unknowns != null)
         {
            Set<HQLAst> createdBooleans = simpifyUnknowns(unknowns);
            if (createdBooleans != null)
            {
               if (needSimplification == null)
               {
                  needSimplification = new HashSet<HQLAst>();
               }
               needSimplification.addAll(createdBooleans);
            }
         }

         // perform simplification of expressions containing boolean operands
         // (see trySimplifyBooleans() function description)
         if (needSimplification != null)
         {
            for (HQLAst next : needSimplification)
            {
               root = trySimplifyBooleans(next, root);
            }
         }
         
         // replace "then ? else ?" with "then cast(? as <datatype>) else
         // cast(? as <datatype>)" for H2 dialect
         if (needExplicitCastInsideTernary != null)
         {
            for (HQLAst next : needExplicitCastInsideTernary)
            {
               makeExplicitCast(next, 1);
               makeExplicitCast(next, 2);
            }
         }
      }
      catch (AstException exc)
      {
         // TODO:  use more specific error number.
         throw new PersistenceException("Error rewriting HQL expression", -1, exc);
      }
      
      if (earlyPublishEntities != null)
      {
         ((ArrayList<String>) earlyPublishEntities).trimToSize();
         earlyPublishEntities = Collections.unmodifiableList(earlyPublishEntities);
      }
      
      return root;
   }

   /**
    * Get rid of unknowns in an HQL AST by evaluating relations in which these
    * unknowns participate.
    *
    * @param  unknowns
    *         Set of HQL nodes which represent all unknowns in the HQL AST.
    *         Nodes can be substitution parameters or whole statements which
    *         evaluate to unknown values.
    *
    * @return Set of HQL <code>BOOL_TRUE</code> and <code>BOOL_FALSE</code>
    *         nodes which were created during the simplifications of unknowns.
    *         May be <code>null</code>.    
    */
   private Set<HQLAst> simpifyUnknowns(Set<HQLAst> unknowns)
   {
      Set<HQLAst> createdBooleans = null;
      Set<HQLAst> skipSet         = null;
      HQLAst next                 = null;
      Iterator<HQLAst> iter       = unknowns.iterator();

      while (true)
      {
         if (next == null)
         {
            if (iter.hasNext())
            {
               next = iter.next();
            }
            else
            {
               break;
            }
         }

         if (skipSet != null && skipSet.contains(next))
         {
            next = null;
            continue;
         }

         HQLAst parent = (HQLAst) next.getParent();

         int disposition = -1;
         if (parent != null)
         {
            int pType              = parent.getType();
            int idx                = next.getIndexPos();
            HQLAst oppositeOperand = (HQLAst) ((idx == 0) ?
                                       next.getNextSibling() :
                                       next.getPrevSibling());

            // Decide what transformation is needed.
            if (oppositeOperand == null)
            {
               if (parent.getType() == IS_NULL)
               {
                  disposition = BOOL_TRUE;
               }
               else if (parent.getType() == NOT_NULL)
               {
                  disposition = BOOL_FALSE;
               }
               else
               {
                  throw new IllegalStateException(
                        "Unexpected usage of standalone unknown value.");
               }
            }
            else if (unknowns.contains(oppositeOperand))
            {
               // comparing two unknowns

               disposition = pType == EQUALS ||
                             pType == GTE ||
                             pType == LTE ? BOOL_TRUE : BOOL_FALSE;

               // skip processing of the opposite unknown operand in the
               // future
               if (skipSet == null)
               {
                  skipSet = new HashSet<HQLAst>();
               }
               skipSet.add(oppositeOperand);
            }
            else if (oppositeOperand.getType() == SUBST ||
                     oppositeOperand.getType() == NUM_LITERAL ||
                     oppositeOperand.getType() == DEC_LITERAL ||
                     oppositeOperand.getType() == BOOL_FALSE ||
                     oppositeOperand.getType() == BOOL_TRUE ||
                     oppositeOperand.getType() == STRING)
            {
               // we have a simple "? <comparison operator> {CONST|SUBST}"
               // statement

               if (pType == EQUALS)
               {
                  disposition = BOOL_FALSE;
               }
               else if (pType == NOT_EQ)
               {
                  disposition = BOOL_TRUE;
               }
               else
               {
                  // results of comparison is unknown value, process it
                  // on the next iteration
                  next = parent;
                  continue;
               }
            }
            else
            {
               // this code correctly handles "field <comparison operator> ?"
               // comparisons, results for more complex statements may be
               // incorrect
               // TODO: handle complex comparisons, e.g. "subst <operator>
               // {const|subst} <comparison operator> ?" 

               boolean nullable = isOppositeOperandNullable(next, idx);

               if (pType == EQUALS ||
                   (pType == GTE && idx == 1) ||
                   (pType == LTE && idx == 0))
               {
                  disposition = (nullable ? IS_NULL : BOOL_FALSE);
               }
               else if (pType == NOT_EQ ||
                        (pType == LT && idx == 1) ||
                        (pType == GT && idx == 0))
               {
                  disposition = (nullable ? NOT_NULL : BOOL_TRUE);
               }
               else if ((pType == GT && idx == 1) ||
                        (pType == LT && idx == 0))
               {
                  disposition = BOOL_FALSE;
               }
               else if ((pType == LTE && idx == 1) ||
                        (pType == GTE && idx == 0))
               {
                  disposition = BOOL_TRUE;
               }
            }
         }
         else
         {
            // top-level unknown value is translated to "false"
            disposition = BOOL_FALSE;
            parent = next;
         }

         // Transform the expression.
         switch (disposition)
         {
            case IS_NULL:
               parent.setType(IS_NULL);
               parent.setText("is null");
               parent.putAnnotation("inlined", Boolean.TRUE);
               next.remove();
               inline = true;
               break;
            case NOT_NULL:
               parent.setType(NOT_NULL);
               parent.setText("is not null");
               parent.putAnnotation("inlined", Boolean.TRUE);
               next.remove();
               inline = true;
               break;
            case BOOL_FALSE:
               parent.setType(BOOL_FALSE);
               parent.setText("false");
               parent.putAnnotation("inlined", Boolean.TRUE);
               parent.removeChildren();
               inline = true;

               if (createdBooleans == null)
               {
                  createdBooleans = new HashSet<HQLAst>();
               }
               createdBooleans.add(parent);
               break;
            case BOOL_TRUE:
               parent.setType(BOOL_TRUE);
               parent.setText("true");
               parent.putAnnotation("inlined", Boolean.TRUE);
               parent.removeChildren();
               inline = true;

               if (createdBooleans == null)
               {
                  createdBooleans = new HashSet<HQLAst>();
               }
               createdBooleans.add(parent);
               break;
            default:
               break;
         }

         next = null;
      }

      return createdBooleans;
   }

   /**
    * Replaces the element of a ternary ("then element1 else element2") with "cast(? as
    * &lt;datatype&gt;)" if it represents a substitution parameter. E.g. the element will be
    * replaced with "cast(? as big_decimal)" is we have a decimal variable.<p>
    * Originally designed as a workaround of the H2 bug.
    *
    * @param ternaryElement
    *        Node which represents the ternary element (i.e. the child of "when" or "then") in
    *        which replacement should be done.
    * @param index
    *        Index of the ternary element in ternary ("1" or "2").
    */
   private void makeExplicitCast(HQLAst ternaryElement, int index)
   {
      HQLAst casted = (HQLAst) ternaryElement.getChildAt(index);

      if (casted == null || casted.getType() != SUBST)
         return;

      String datatype = (String) casted.getAnnotation("datatype");
      datatype = getDatatypeSubstitutionForCastFunction(datatype);

      HQLAst castFunc = new HQLAst();
      castFunc.setType(CAST);
      castFunc.setText("cast");

      HQLAst castedType = new HQLAst();
      castedType.setType(SYMBOL);
      castedType.setText(datatype);

      casted.remove();
      ternaryElement.graftAt(castFunc, index);
      castFunc.graft(casted);
      castFunc.graft(castedType);
   }

   /**
    * Returns for the given HQL type its replacement which can be used into
    * CAST function.
    *
    * @param  datatype
    *         Original HQL type.
    *
    * @return Replacement of the given HQL type with can be used into CAST
    *         function.
    */
   private String getDatatypeSubstitutionForCastFunction(String datatype)
   {
      if (datatype == null)
      {
         throw new IllegalArgumentException("No datatype specified.");
      }

      HQLTypes type = HQLTypes.valueOf(datatype);
      switch (type)
      {
         case BOOLEAN:
            return "boolean";
         case DATE:
            return "date";
         case DATETIME:
            return "timestamp";
         case DATETIMETZ:
            return "timestamp"; //TODO not sure if this is correct
         case DECIMAL:
            return "big_decimal";
         case INTEGER:
            return "integer";
         case RAW:
            return "binary";
         case TEXT:
            return "string";
         case DOUBLE:
            return "double";
         case LONG:
            return "long";
         default:
            throw new IllegalArgumentException(
               "Cannot find replacement for the '" + datatype + "'.");
      }
   }

   /**
    * Tries to roll up subexpressions according to the following rules:
    * <ul>
    *   <li>{condition} or false --> {condition}</li>
    *   <li>{condition} or  true  --> true</li>
    *   <li>{condition} and true  --> {condition}</li>
    *   <li>{condition} and false --> false</li>
    * </ul><p>
    * If an expression has been rolled up to a boolean value, it tries to roll
    * this value up with other conditions and so on.
    *
    * @param  booleanOperand
    *         A node which represents a boolean operand in the expressions
    *         above.
    * @param  root
    *         The root node of the whole HQL tree.
    *
    * @return The new root node of the whole HQL tree (the root node may be
    *         replaced if we have rolled up the first-level leaves).
    *
    * @throws AstException
    *         If there is any error grafting subtrees in the process of
    *         restructuring the subtree.
    */
   private HQLAst trySimplifyBooleans(HQLAst booleanOperand, HQLAst root)
   throws AstException
   {
      // check that the brunch which contains this operand wasn't thrown away
      // during the processing before
      Aast par = booleanOperand.getParent();
      boolean valid = false;
      while (par != null)
      {
         if (par == root)
         {
            valid = true;
            break;
         }
         else
            par = par.getParent();
      }
      
      if (!valid)
         return root;

      HQLAst parent       = (HQLAst) booleanOperand.getParent();
      int type            = booleanOperand.getType();
      int pType           = parent.getType();
      HQLAst otherOperand = (HQLAst) (booleanOperand.getIndexPos() == 0
                                       ? booleanOperand.getNextSibling()
                                       : booleanOperand.getPrevSibling());
      boolean newBooleanElementAdded = false;

      // note that in the processing below, when we replace the node we break
      // all links with old nodes in both directions in order to ensure that  
      // the detection of detached branches above works correctly
      switch (pType)
      {
         case AND:
            switch (type)
            {
               case BOOL_TRUE:
                  Aast operandParent = parent.getParent();
                  if (operandParent == null)
                  {
                     for (int i = 0; i < root.getNumberOfChildren(); i++)
                     {
                        root.getChildAt(i).setParent(null);
                     }

                     root = otherOperand;
                  }
                  else
                  {
                     int operandIndex = parent.getIndexPos();
                     operandParent.graftAt(otherOperand, operandIndex);
                     parent.remove();
                     parent.setParent(null);
                  }
                  break;
               case BOOL_FALSE:
                  parent.setType(BOOL_FALSE);
                  parent.setText("false");
                  parent.putAnnotation("inlined", Boolean.TRUE);

                  for (int i = 0; i < parent.getNumberOfChildren(); i++)
                  {
                     parent.getChildAt(i).setParent(null);
                  }
                  parent.removeChildren();
                  inline = true;
                  newBooleanElementAdded = true;
                  break;
            }
            break;
         case OR:
            switch (type)
            {
               case BOOL_TRUE:
                  parent.setType(BOOL_TRUE);
                  parent.setText("true");
                  parent.putAnnotation("inlined", Boolean.TRUE);

                  for (int i = 0; i < parent.getNumberOfChildren(); i++)
                  {
                     parent.getChildAt(i).setParent(null);
                  }
                  parent.removeChildren();
                  inline = true;
                  newBooleanElementAdded = true;
                  break;
               case BOOL_FALSE:
                  Aast operandParent = parent.getParent();
                  if (operandParent == null)
                  {
                     for (int i = 0; i < root.getNumberOfChildren(); i++)
                     {
                        root.getChildAt(i).setParent(null);
                     }

                     root = otherOperand;
                  }
                  else
                  {
                     int operandIndex = parent.getIndexPos();
                     operandParent.graftAt(otherOperand, operandIndex);
                     parent.remove();
                     parent.setParent(null);
                  }
                  break;
            }
            break;
      }

      // perform recursive roll up
      if (newBooleanElementAdded)
         root = trySimplifyBooleans(parent, root);

      return root;
   }

   /**
    * Replace legacy indexed property with extent with custom property. Mark ast as inlined.
    *
    * @param   ast
    *          A node which represents a denormalized field with extent.
    * @param   parameters
    *          Query substitution parameters for the where clause.
    */
   private void inlineDenormalizedField(HQLAst ast, Object[] parameters)
   {
      HQLAst subst = (HQLAst) ast.getChildAt(0).getChildAt(0);

      int extentIndex;
      switch (subst.getType())
      {
         case SUBST:
            String dataType = (String) subst.getAnnotation("datatype");
            int index = ((Long) subst.getAnnotation("index")).intValue();
            Object parm = parameters[index];

            HQLTypes hqlType = Enum.valueOf(HQLTypes.class, dataType);
            switch (hqlType)
            {
               case INTEGER:
                  extentIndex = ((integer) parm).intValue();
                  break;
               case LONG:
                  extentIndex = ((int64) parm).intValue();
                  break;
               default:
                  return;
            }
            break;
         case NUM_LITERAL:
            extentIndex = Integer.parseInt(subst.getText());
            break;
         default:
            return;
      }

      String denormalizedProperty = TableMapper.getDenormalizedProperty(dmoClass,
                                                                        ast.getText(),
                                                                        extentIndex);
      ast.setText(denormalizedProperty);
      ast.removeChildren();
      ast.putAnnotation("inlined", Boolean.TRUE);
      inline = true;
   }

   /**
    * Attempt to inline a single query substitution parameter, if it is a
    * numeric, text, or date type.  This involves retasking a substitution
    * parameter node to a be a different type, and changing its text, so that
    * the original placeholder (<code>?</code>) is replaced with the text of
    * the parameter's value.  The node is annotated as "inlined" to enable
    * downstream processing to accurately track the remaining placeholders.
    * 
    * @param   ast
    *          AST node which represents the query substitution parameter to
    *          be inlined.
    * @param   parameters
    *          The array of substitution parameters from which the value of
    *          the inlined parameter will be determined.
    * @param   dialect
    *          Database dialect in use.
    *
    * @return  The HQL type of the inlined parameter, or <code>SUBST</code> if
    *          the parameter wasn't inlined.
    */
   private int inlineSubstitutionParameter(HQLAst ast,
                                           Object[] parameters,
                                           P2JDialect dialect)
   {
      String text = null;
      String dataType = (String) ast.getAnnotation("datatype");
      int index = ((Long) ast.getAnnotation("index")).intValue();
      Object parm = parameters[index];
      if (parm instanceof Resolvable && !(parm instanceof FieldReference))
      {
         // Do not inline Resolvables.  If we get here, it represents a
         // programming error.  Warn, but don't fail.
         if (LOG.isLoggable(Level.SEVERE))
         {
            StringBuilder buf = new StringBuilder("Resolvable substitution parameter @");
            buf.append(index);
            buf.append(" (");
            buf.append(parm);
            buf.append(") cannot be inlined [");
            buf.append(hql);
            buf.append("]");
            LOG.log(Level.SEVERE, buf.toString(), new Throwable());
         }
         
         return SUBST;
      }
      
      HQLTypes hqlType = Enum.valueOf(HQLTypes.class, dataType);
      int tokType = SUBST;
      
      switch (hqlType)
      {
         case INTEGER:
         case LONG:
            tokType = NUM_LITERAL;
            break;
            
         case DECIMAL:
         case DOUBLE:
            tokType = DEC_LITERAL;
            break;
            
         case DATE:
            parm = Persistence.preprocessQueryParameter(parm, dialect);
            tokType = STRING;
            text = dialect.formatDate((date) parm);
            break;
            
         case TEXT:
            parm = Persistence.preprocessQueryParameter(parm, dialect);
            text = ((character) parm).toStringMessage();
            if (ast.getType() != LIKE || text.endsWith("%"))
            {
               tokType = STRING;
            }
            text = character.postprocessStringLiteral(text, true);
            break;

         case BOOLEAN:
            if (((logical) parm).booleanValue())
            {
               tokType = BOOL_TRUE;
               text = "true";
            }
            else
            {
               tokType = BOOL_FALSE;
               text = "false";
            }
            break;

         //TODO handle DATETIME and DATETIMETZ
            
         default:
            // Do not inline any other type.
            break;
      }
      
      if (tokType != SUBST)
      {
         if (text == null)
         {
            if (parm instanceof BaseDataType)
            {
               text = ((BaseDataType) parm).toStringMessage();
            }
            else
            {
               text = parm.toString();
            }
         }
         else if (tokType == STRING)
         {
            // Enclose text in single quotes.
            StringBuilder buf = new StringBuilder("'");
            buf.append(text);
            buf.append("'");
            text = buf.toString();
         }
         
         ast.setText(text);
         ast.setType(tokType);
         ast.putAnnotation("inlined", Boolean.TRUE);
         inline = true;
      }

      return tokType;
   }
   
   /**
    * Manually disambiguate a user defined function by taking its overloaded
    * function name and data type signature and replacing the name in its AST
    * with a unique name generated when the function was registered.
    * 
    * @param   function
    *          AST node for the function.
    * 
    * @throws  PersistenceException
    *          if there is an error determining the data type of any parameter
    *          of the function.
    */
   private void manuallyOverload(HQLAst function)
   throws PersistenceException
   {
      String name = function.getText();
      int len = function.getNumImmediateChildren();
      HQLTypes[] signature = new HQLTypes[len];
      for (int i = 0; i < len; i++)
      {
         Aast child = function.getChildAt(i);
         signature[i] = DataTypeHelper.expressionType(child, dmoClass);
      }
      
      FunctionKey key = new FunctionKey(database, name, signature);
      String newName = overloadedFunctions.get(key);
      
      if (LOG.isLoggable(Level.FINEST))
      {
         LOG.log(Level.FINEST, "Overloading function '" + key + "' with '" + newName + "'");
      }
      
      if (newName != null)
      {
         function.setText(newName);
      }
      else if (LOG.isLoggable(Level.FINE))
      {
         LOG.log(Level.FINE, "Unable to overload function [" + key + "]");
         if (LOG.isLoggable(Level.FINEST))
         {
            StringBuilder buf = new StringBuilder("Available overloads:\n");
            buf.append(overloadedFunctions);
            LOG.log(Level.FINEST, buf.toString());
         }
      }
   }
   
   /**
    * Given an AST which represents the alias qualifier of a text property,
    * replace the property node's text (the property name) with a name which
    * represents a computed column.  The computed column is a mechanism used
    * by some databases to embed expressions into a synthetic column,
    * particularly for use in index definitions.
    * <p>
    * If the property reference is already embedded within an
    * <code>upper()</code> function call, that call is discarded if it is
    * redundant with the computed column definition.
    * 
    * @param   alias
    *          Alias AST node.
    * 
    * @return  <code>true</code> if the computed column reference was
    *          injected;  <code>false</code> if the current database dialect
    *          does not use computed columns, or if the given property is not
    *          a text property.
    * 
    * @throws  AstException
    *          if there is an error restructuring the alias node or its parent
    *          node.
    */
   private boolean injectComputedColumn(HQLAst alias)
   throws AstException
   {
      RecordBuffer buffer = resolveBuffer(alias);
      P2JDialect dialect = null;
      if (database.isDirty())
      {
         dialect = (P2JDialect) DatabaseManager.getDialect(database);
      }
      else
      {
         dialect = buffer.getDialect();
      }
      if (!dialect.needsComputedColumns())
      {
         return false;
      }
      
      HQLAst propAst = (HQLAst) alias.getFirstChild();
      String property = propAst.getText();
      Boolean ignoreCase = DatabaseManager.getIgnoreCase(dmoClass, property);
      if (ignoreCase == null)
      {
         return false;
      }
      HQLAst parent = (HQLAst) alias.getParent();
      
      // If the computed column is designed to ignore case (i.e., has an
      // embedded upper() function) AND the alias already is contained within
      // its own upper() function in the tree, we need to replace the upper()
      // function node with the alias node.
      if (ignoreCase &&
          parent.getType() == FUNCTION &&
          "upper".equalsIgnoreCase(parent.getText()))
      {
         int idxPos = parent.getIndexPos();
         HQLAst grandparent = (HQLAst) parent.getParent();
         parent.remove();
         grandparent.graftAt(alias, idxPos);
      }
      
      // Prepend the computed column prefix to the property name.
      propAst.setText(dialect.getComputedColumnPrefix(!ignoreCase) + property);
      return true;
   }
   
   /**
    * Work around a defect in the Hibernate HQL parser which requires that an
    * expression which returns a boolean always be an operand in a binary
    * comparison test.  This requires that simple where clauses such as
    * <code>where true</code> be rewritten as <code>where true = true</code>,
    * and a case statement such as <code>case when ? then ? else ? end</code>
    * be rewritten as <code>case when ? = true then ? else ? end</code>.
    *
    * @param   root
    *          Root node of the AST for the HQL expression.
    *
    * @return  The root node of the AST for the HQL expression, which may be
    *          a different node than the root passed in.
    *
    * @throws  PersistenceException
    *          if any error occurred during rewriting the AST.
    */
   /* See header entry #008
   private HQLAst preprocessBooleanExpressions(HQLAst root)
   throws PersistenceException
   {
      Set targets = null;
      
      // Walk the tree and identify logical conditions which are not parented
      // by a comparison operator.  Rewrite them so that they emit as:
      // <expr> = true.
      Iterator iter = root.iterator();
      while (iter.hasNext())
      {
         HQLAst next = (HQLAst) iter.next();
         HQLAst target = null;
         
         switch (next.getType())
         {
            case BOOL_TRUE:
            case BOOL_FALSE:
            case SUBST:
            case FUNCTION:
               HQLAst parent = (HQLAst) next.getParent();
               if (parent == null)
               {
                  target = next;
               }
               else
               {
                  switch (parent.getType())
                  {
                     case AND:
                     case OR:
                     case NOT:
                        target = next;
                        break;
                     case TERNARY:
                        if (next.getIndexPos() == 0)
                        {
                           target = next;
                        }
                        break;
                     default:
                        break;
                  }
               }
               break;
               
            default:
               break;
         }
         
         if (target != null)
         {
            if (targets == null)
            {
               targets = new HashSet();
            }
            targets.add(next);
         }
      }
      
      if (targets != null)
      {
         try
         {
            iter = targets.iterator();
            while (iter.hasNext())
            {
               HQLAst next = (HQLAst) iter.next();
               HQLAst parent = (HQLAst) next.getParent();
               
               HQLAst equals = new HQLAst();
               equals.setType(EQUALS);
               equals.setText("=");
               
               HQLAst boolTrue = new HQLAst();
               boolTrue.setType(BOOL_TRUE);
               boolTrue.setText("true");
               
               if (parent == null)
               {
                  root = parent = equals;
               }
               else
               {
                  int idxPos = next.getIndexPos();
                  next.remove();
                  parent.graftAt(equals, idxPos);
               }
               
               equals.graft(next);
               equals.graft(boolTrue);
            }
         }
         catch (AstException exc)
         {
            // TODO:  use more specific error number.
            throw new PersistenceException("Error rewriting HQL expression",
                                           -1,
                                           exc);
         }
      }
      
      return root;
   }
   */
   
   /**
    * Analyze the AST generated for the HQL expression, identifying those
    * subtrees which are targets for conversion.  A subtree is a target if
    * it is an alias whose first child is a property reference with a
    * subscript, AND the alias/property combination is associated with a
    * composite element in the Hibernate configuration.
    * <p>
    * The AST subtree structure which represents our target is as follows:
    * <pre>
    *    ALIAS
    *    |
    *    +--PROPERTY
    *       |
    *       +--LBRACKET
    *          |
    *          +--{...} (SUBST or NUM_LITERAL)
    * </pre>
    * <p>
    * In its original form, this would look like:
    * <pre>
    *    alias.property[subscript]
    * </pre>
    * <p>
    * If any such constructs are identified, they are added to a map of
    * targets and returned.
    *
    * @param   root
    *          Root node of the AST for the HQL expression.
    *
    * @return  Map of target AST subtrees to composite names, or
    *          <code>null</code> if no targets were found.  The composite
    *          names are retrieved by the {@link DatabaseManager} from the
    *          appropriate Hibernate ORM configurations.
    */
   private Map<HQLAst, String> analyzeComposites(HQLAst root)
   {
      Map<HQLAst, String> targets = null;
      
      Iterator<Aast> iter = root.iterator();
      while (iter.hasNext())
      {
         HQLAst next = (HQLAst) iter.next();
         
         HQLAst alias = null;
         HQLAst property = null;
         HQLAst subscript = null;
         
         Class<?> dmoClass = null;
         
         // Do we have an alias node?
         if (next.getType() == ALIAS)
         {
            alias = next;
            property = (HQLAst) next.getChildAt(0);
            
            // Does the alias have a property child?
            if (property != null && property.getType() == PROPERTY)
            {
               // Look up DMO implementation class which corresponds with
               // alias.
               RecordBuffer buffer = resolveBuffer(alias);
               dmoClass = buffer.getDMOImplementationClass();
               String entity = buffer.getEntityName().intern();
               
               // At this point, we have a DMO entity and a property.  Though
               // not part of the composite analysis, this is a convenient
               // place to store off our restriction property names without
               // requiring a separate tree walk.
               addRestrictionProperty(entity, property.getText());
               
               // Look for an LBRACKET, which indicates a possible composite.
               HQLAst lbracket = (HQLAst) property.getChildAt(0);
               
               // Does the property have a subscript?
               if (lbracket != null && lbracket.getType() == LBRACKET)
               {
                  subscript = (HQLAst) lbracket.getChildAt(0);
               }
            }
            else
            {
               property = null;
            }
         }
         else
         {
            alias = null;
         }
         
         if (alias == null || property == null || subscript == null)
         {
            continue;
         }
         
         // Check Hibernate configuration to see if we match a composite for
         // this alias and property combination.
         String propName = property.getText();
         String composite = DatabaseManager.getComposite(dmoClass, propName);
         
         // Add composite entry to the map.
         if (composite != null)
         {
            if (targets == null)
            {
               targets = new HashMap<HQLAst, String>();
            }
            
            targets.put(alias, composite);
         }
      }
      
      return targets;
   }

   /**
    * Get list of denormalized fields with extent for where clause AST.
    *
    * @param   root
    *          Root node of the AST for the HQL expression.
    *
    * @return  List of AST of denormalized fields with extent.
    */
   private Iterable<HQLAst> getDenormalizedFields(HQLAst root)
   {
      List<HQLAst> denormalizedFields = null;

      Iterator<Aast> iter = root.iterator();
      while (iter.hasNext())
      {
         HQLAst next = (HQLAst) iter.next();

         HQLAst alias;
         HQLAst property = null;
         HQLAst subscript = null;

         // Do we have an alias node?
         if (next.getType() == ALIAS)
         {
            alias = next;
            property = (HQLAst) next.getChildAt(0);

            // Does the alias have a property child?
            if (property != null && property.getType() == PROPERTY)
            {
               // Look up DMO implementation class which corresponds with
               // alias.
               RecordBuffer buffer = resolveBuffer(alias);
               Class<?> dmoClass = buffer.getDMOImplementationClass();

               if (TableMapper.getLegacyFieldInfo(dmoClass, property.getText()) == null)
               {
                  // Look for an LBRACKET, which indicates a possible composite.
                  HQLAst lbracket = (HQLAst) property.getChildAt(0);

                  // Does the property have a subscript?
                  if (lbracket != null && lbracket.getType() == LBRACKET)
                  {
                     subscript = (HQLAst) lbracket.getChildAt(0);
                  }
               }
               else
               {
                  property = null;
               }
            }
            else
            {
               property = null;
            }
         }
         else
         {
            alias = null;
         }

         if (alias == null || property == null || subscript == null)
         {
            continue;
         }

         if (denormalizedFields == null)
         {
            denormalizedFields = new ArrayList<>();
         }

         denormalizedFields.add(property);
      }

      return denormalizedFields;
   }

   /**
    * Add the given property name to the set of properties used by this where
    * clause as restriction criteria.  Each set of properties is mapped to
    * the name of the DMO entity with which the property is associated.
    * <p>
    * In the current implementation, the nature of the restriction is not
    * preserved (e.g., equality match, function parameter, etc.).
    *
    * @param   entity
    *          DMO entity containing the given property.
    * @param   property
    *          Name of a property used by this where clause as a restrictive
    *          criterion.
    */
   private void addRestrictionProperty(String entity, String property)
   {
      if (restrictionProperties == null)
      {
         restrictionProperties = new HashMap<String, Set<String>>();
      }
      
      Set<String> props = restrictionProperties.get(entity);
      if (props == null)
      {
         props = new HashSet<String>();
         restrictionProperties.put(entity, props);
      }
      
      props.add(property);
   }
   
   /**
    * Restructure the AST subtrees (if any) defined as rewrite targets by the
    * {@link #analyzeComposites composite analysis step}.  This involves the
    * expansion of an expression, for instance:...
    * <pre>
    *    OR [ or ]
    *    |
    *    +--EQUALS [ = ]
    *    |  |
    *    |  +--ALIAS [ alias ]
    *    |  |  |
    *    |  |  +--PROPERTY [ propertyABC ]
    *    |  |     |
    *    |  |     +--LBRACKET [ [ ]
    *    |  |        |
    *    |  |        +--NUM_LITERAL [ 0 ]
    *    |  |
    *    |  +--SUBST [ ? ]
    *    |
    *    +--EQUALS [ = ]
    *       |
    *       +--ALIAS [ alias ]
    *       |  |
    *       |  +--PROPERTY [ propertyABC ]
    *       |     |
    *       |     +--LBRACKET [ [ ]
    *       |        |
    *       |        +--NUM_LITERAL [ 1 ]
    *       |
    *       +--SUBST [ ? ]
    * </pre>
    * ...to an expresion which consists of the modified original expression
    * and index specifications added to the lop level of the tree using
    * logical AND operators:
    * <pre>
    *   AND [ and ]
    *   |
    *   +--AND [ and ]
    *   |  |
    *   |  +--LPARENS [ ( ]
    *   |  |  |
    *   |  |  +--OR [ or ]
    *   |  |     |
    *   |  |     +--EQUALS [ = ]
    *   |  |     |  |
    *   |  |     |  +--ALIAS [ alias_composite2_0 ]
    *   |  |     |  |  |
    *   |  |     |  |  +--PROPERTY [ propertyABC ]
    *   |  |     |  |
    *   |  |     |  +--SUBST [ ? ]
    *   |  |     |
    *   |  |     +--EQUALS [ = ]
    *   |  |        |
    *   |  |        +--ALIAS [ alias_composite2_1 ]
    *   |  |        |  |
    *   |  |        |  +--PROPERTY [ propertyABC ]
    *   |  |        |
    *   |  |        +--SUBST [ ? ]
    *   |  |
    *   |  +--EQUALS [ = ]
    *   |     |
    *   |     +--FUNCTION [ index ]
    *   |     |   |
    *   |     |   +--ALIAS [ alias_composite2_0 ]
    *   |     |
    *   |     +--NUM_LITERAL [ 0 ]
    *   |
    *   +--EQUALS [ = ]
    *      |
    *      +--FUNCTION [ index ]
    *      |   |
    *      |   +--ALIAS [ alias_composite2_1 ]
    *      |
    *      +--NUM_LITERAL [ 1 ]
    * </pre>
    * <p>
    * This corresponds to the following expression expansion (for this
    * expression, two composite targets will be passed to the function):
    * <pre>
    *    alias.propertyABC[0] = ? or alias.propertyABC[1] = ?
    * </pre>
    * ...to this one:
    * <pre>
    *    (alias_composite2_0.propertyABC = ? or
    *      alias_composite2_1.propertyABC = ?) and
    *      index(alias_composite2_0) = 0 and
    *      index(alias_composite2_1) = 1
    * </pre>
    * However, the infix example above is informational only;  the generation
    * of the final, infix form of the expression is performed in a later step
    * (see {@link #emit}).
    *
    * @param   targets
    *          Map of AST subtrees targeted for rewriting by the analysis
    *          step.  The map's keys are the subtrees themselves;  the map's
    *          values are the names of the composite list entities retrieved
    *          by the {@link DatabaseManager}.
    * @param   root
    *          The root node of the AST of the whole HQL expression.           
    *
    * @return  The root node of the restructured ASTs.
    *
    * @throws  PersistenceException
    *          if there is any error restructuring the ASTs.
    */
   private HQLAst restructure(HQLAst root, Map<HQLAst, String> targets)
   throws PersistenceException
   {
      try
      {
         // round-bracket the whole original expression
         if (root.getType() != LPARENS)
         {
            HQLAst lparens = new HQLAst();
            lparens.setText("(");
            lparens.setType(LPARENS);
            lparens.graft(root);
         }

         HQLAst last = null;
         Iterator<Map.Entry<HQLAst, String>> iter =
            targets.entrySet().iterator();
         while (iter.hasNext())
         {
            Map.Entry<HQLAst, String> next = iter.next();
            HQLAst alias = next.getKey();
            String compName = next.getValue();
            rewriteAlias(alias, compName);
            last = alias;
         }
         
         return (HQLAst) last.getAncestor(-1);
      }
      catch (AstException exc)
      {
         // TODO:  use more specific error number.
         throw new PersistenceException("Error rewriting HQL expression",
                                        -1,
                                        exc);
      }
   }
   
   /**
    * Given a particular, target subtree and a composite list entity name,
    * rewrite the AST as described in the {@link #restructure} method.
    *
    * @param   alias
    *          Alias subtree to be expanded.
    * @param   compName
    *          Name of the composite list associated with the combination of
    *          the alias name and property name encountered in the
    *          <code>alias</code> subtree.
    *
    * @throws  AstException
    *          if there is any error grafting subtrees in the process of
    *          restructuring the subtree.
    */
   private void rewriteAlias(HQLAst alias, String compName)
   throws AstException
   {
      // Get the ASTs we need to manipulate.
      HQLAst root = (HQLAst) alias.getAncestor(-1);
      HQLAst property = (HQLAst) alias.getChildAt(0);
      HQLAst subscript = (HQLAst) property.getChildAt(0).getChildAt(0);
      String subIndex = (subscript.getType() == SUBST
                         ? String.valueOf(subscript.getAnnotation("index"))
                         : subscript.getText());
      
      // Create qualified name of composite alias.
      StringBuilder buf = new StringBuilder(alias.getText());
      buf.append("_");
      buf.append(compName);
      buf.append("_");
      buf.append(subIndex);
      String verboseName = buf.toString();
      
      // Reset alias AST's text to the composite name;  store original name
      // in an annotation;  we will need it when generating join clauses for
      // this composite.
      alias.putAnnotation("original_name", alias.getText());
      alias.setText(verboseName);
      
      // Create the AND operator node to accommodate the second half of the
      // test (the index check).
      HQLAst and = new HQLAst();
      and.setType(AND);
      and.setText("and");
      and.graft(root);
      
      // Create the equality operator for the index check as the second
      // child of the AND.
      HQLAst equals = new HQLAst();
      equals.setType(EQUALS);
      equals.setText("=");
      and.graft(equals);
      
      // Create a function node for the HQL index function.
      HQLAst index = new HQLAst();
      index.setType(FUNCTION);
      index.setText("index");
      equals.graft(index);
      
      // Create a parameter node for the index function, using the new
      // composite alias name.
      HQLAst composite = new HQLAst();
      composite.setType(ALIAS);
      composite.setText(verboseName);
      index.graft(composite);
      
      // Add a copy of the original subscript as the second child of the
      // index check equality operator.
      equals.graft(subscript.duplicateFresh());
      
      // Remove subscript from its original location.
      subscript.getParent().remove();
   }
   
   /**
    * Emit the {@link #restructure restructured} where clause AST into an
    * infix notation expression, which is compliant with the HQL syntax
    * supported by Hibernate.  This is the post-rewrite, "finished product"
    * expression which will be returned by the {@link #getHQL} method.
    *
    * @param   root
    *          Root of the restructured where clause AST.
    *
    * @return  Rewritten HQL where clause subexpression.
    */
   private HQLExpression emit(HQLAst root)
   {
      final HQLExpression buf = new HQLExpression();
      
      // Define a walk listener which will inject content into the output
      // string buffer during appropriate tree walk events.
      AstWalkListener awl = new AstWalkListener()
      {
         /**
          * Called when the AST walk ascends between levels of the tree. Used
          * to close off certain constructs when emitting HQL.
          *
          * @param   ast
          *          AST node to which the walk is ascending.
          */
         public void ascent(Aast ast)
         {
            if (ast == null)
            {
               return;
            }
            
            switch (ast.getType())
            {
               case LPARENS:
               case FUNCTION:
               case CAST:
                  buf.append(")");
                  break;
                  
               case TERNARY:
                  buf.append(") end");
                  break;
                  
               case LBRACKET:
                  buf.append("]");
                  break;
                  
               case IS_NULL:
                  buf.append(" is null");
                  break;
                  
               case NOT_NULL:
                  buf.append(" is not null");
                  break;
                  
               case SELECT:
               case FROM:
                  buf.append(" ");
                  break;
                  
               default:
                  break;
            }
         }
         
         /**
          * Called when the AST walk descends between levels of the tree.
          * Used to emit a dot operator after an alias, in order to
          * dereference a property.
          *
          * @param   ast
          *          AST node from which the walk is descending.
          */
         public void descent(Aast ast)
         {
            if (ast == null)
            {
               return;
            }
            
            switch (ast.getType())
            {
               case ALIAS:
                  buf.append(".");
                  break;
                  
               case SELECT:
               case FROM:
               case WHERE:
                  buf.append(" ");
                  break;
                  
               default:
                  break;
            }
         }
         
         /**
          * Called when the AST walk moves laterally between sibling nodes.
          * Used to emit the symbol of a binary operator between operands of
          * a binary operation, or to emit a comma between function
          * parameters.
          *
          * @param   ast
          *          AST parent node of the siblings between which the walk
          *          is moving laterally.
          * @param   index
          *          Zero-based index of the sibling node about to be visited.
          */
         public void nextChild(Aast ast, int index)
         {
            if (ast == null)
            {
               return;
            }
            
            if (isBinaryOperator(ast))
            {
               buf.append(" ");
               buf.append(ast.getText());
               buf.append(" ");
            }
            else
            {
               switch (ast.getType())
               {
                  case FUNCTION:
                     buf.append(", ");
                     break;

                  case CAST:
                     buf.append(" as ");
                     break;
                     
                  case TERNARY:
                     switch (index)
                     {
                        case 1:
                           buf.append(" then (");
                           break;
                        case 2:
                           buf.append(") else (");
                           break;
                        default:
                           break;
                     }
                     break;
                     
                  case FROM:
                     buf.append(" as ");
                     break;
                     
                  default:
                     break;
               }
            }
         }
      };
      
      // Walk the full tree and emit the expression into the string buffer.
      Iterator<Aast> iter = root.iterator(0, awl);
      int posIndex = 0;
      while (iter.hasNext())
      {
         HQLAst next = (HQLAst) iter.next();
         int ttype = next.getType();
         
         if (!isBinaryOperator(next))
         {
            String text = next.getText();
            switch (ttype)
            {
               case SUBST:
                  Long paramCount = (Long) next.getAnnotation("positional-params"); 
                  if (paramCount == null || paramCount < 2)
                  {
                     buf.append("", true);
                  }
                  else 
                  {
                     buf.append(":px").append(Integer.toString(++posIndex));
                  }
                  break;
                  
               case FUNCTION:
               case CAST:
                  buf.append(text);
                  buf.append("(");
                  if (next.getNumImmediateChildren() == 0)
                  {
                     buf.append(")");
                  }
                  break;
                  
               case TERNARY:
                  buf.append("case when ");
                  break;
                  
               case SUBSELECT:
               case IS_NULL:
               case NOT_NULL:
                  break;

               default:
                  buf.append(text);
                  break;
            }
         }
      }
      
      return buf;
   }
   
   /**
    * Define the set of ANSI-style join statements required to join the
    * "master" DMO with one or more associated lists of composite elements.
    * These will be integrated by the enclosing query into the overall HQL
    * statement.
    * <p>
    * Duplicates are eliminated.  The order in which joins are detected is
    * preserved in the resulting data structure, such that later iterations
    * over the set will return these in the same order in which they were
    * added.
    *
    * @param   targets
    *          Map of AST subtrees targeted for rewriting by the analysis
    *          step.  The map's keys are the subtrees themselves;  the map's
    *          values are the names of the composite list entities retrieved
    *          by the {@link DatabaseManager}.
    *
    * @return  Set of join subexpression strings.  May be empty if no targets
    *          were specified, but will not be <code>null</code>.
    */
   private LinkedHashSet<String> generateJoins(Map<HQLAst, String> targets)
   {
      LinkedHashSet<String> joins = new LinkedHashSet<String>();
      Iterator<Map.Entry<HQLAst, String>> iter =
         targets.entrySet().iterator();
      while (iter.hasNext())
      {
         Map.Entry<HQLAst, String> next = iter.next();
         HQLAst alias = next.getKey();
         String compName = next.getValue();
         String aliasName = (String) alias.getAnnotation("original_name");
         String verboseName = alias.getText();
         joins.add(generateJoin(compName, aliasName, verboseName));
      }
      
      return joins;
   }
   
   /**
    * Generate a single HQL subexpression for an ANSI-style join, which joins
    * a "master" DMO record with an associated list of composite elements.
    *
    * @param   compName
    *          Name of the associated list of composite elements.
    * @param   alias
    *          Alias for the "master" DMO record to which the composite list
    *          is to be joined.
    * @param   verboseAlias
    *          Verbose alias name, including the composite portion and
    *          subscript qualifier.
    * 
    * @return  Join text snippet.
    */
   private String generateJoin(String compName,
                               String alias,
                               String verboseAlias)
   {
      StringBuilder buf = new StringBuilder("join ");
      buf.append(alias);
      buf.append(".");
      buf.append(compName);
      buf.append(" as ");
      buf.append(verboseAlias);
      
      return buf.toString();
   }
   
   /**
    * Create the mapping of substitution parameters to their corresponding
    * substitution placeholders in the rewritten HQL where clause.  These
    * placeholders (<code>?</code>) may have been reordered or removed (due to
    * inlining) during the rewrite process.
    * <p>
    * During parsing of the HQL clause, an index annotation was added to each
    * AST of type <code>SUBST</code> (token type for the query substitution
    * parameter placeholder), denoting that placeholder's position in the
    * original where clause.  We now walk the rewritten AST, extracting this
    * annotation and storing it in a new list.  The resulting array defines
    * the new order of the substitution parameters.  For parameters which
    * were dropped, due to inlining, a null placeholder will appear in the new
    * list.
    *
    * @param   root
    *          Root node of the rewritten where clause AST.
    * @param   parameters
    *          Query substitution parameters for the where clause.
    *
    * @return  Object which maps new query substitution parameter positions to
    *          their positions in the original parameters array.  These
    *          indices are zero-based.
    */
   private ParameterIndices createParameterIndices(HQLAst root, Object[] parameters)
   {
      if (parameters == null)
      {
         return (new ParameterIndices(null));
      }
      
      Integer[] indices = new Integer[parameters.length];
      
      // Extract the "index" annotation from each SUBST node, in order of
      // the tree walk, which will correspond to the order of the placeholders
      // in the final, infix expression.
      int counter = 0;
      Iterator<Aast> iter = root.iterator();
      while (iter.hasNext())
      {
         Long index = null;
         HQLAst next = (HQLAst) iter.next();
         switch (next.getType())
         {
            case SUBST:
               index = (Long) next.getAnnotation("index");
               indices[counter++] = new Integer(index.intValue());
               break;
               
            case NUM_LITERAL:
            case DEC_LITERAL:
            case STRING:
            case IS_NULL:
            case NOT_NULL:
            case BOOL_FALSE:
               if (inline && next.isAnnotation("inlined"))
               {
                  // Allow null to remain at this position in the array.
                  counter++;
               }
               break;
               
            default:
               break;
         }
      }
      
      return (new ParameterIndices(indices));
   }
   
   /**
    * Convenience method to detect whether an AST represents a binary
    * operator, based on its token type.
    *
    * @param   ast
    *          AST node to test.
    *
    * @return  <code>true</code> if the AST's token type indicates a binary
    *          operator, else <code>false</code>.
    */
   private boolean isBinaryOperator(Aast ast)
   {
      switch (ast.getType())
      {
         case OR:
         case AND:
         case EQUALS:
         case NOT_EQ:
         case LIKE:
         case GT:
         case LT:
         case GTE:
         case LTE:
         case PLUS:
         case MINUS:
         case MULTIPLY:
         case DIVIDE:
         case CONCAT:
            return true;
      }
      
      return false;
   }
   
   /**
    * Given an AST and its index, and assuming it is one of two operands in a
    * binary comparison, indicate whether the opposite operand is nullable.
    * It is considered nullable if it is not a non-nullable DMO property.
    * <p>
    * TODO:  currently, we only test if the property is the reserved, primary
    * key property.  We should augment this test to check for other
    * non-nullable properties.
    * 
    * @param   ast
    *          The operand opposite the operand we wish to test.
    * @param   index
    *          The index of <code>ast</code> among its siblings.  This must be
    *          0 or 1, since the expression is binary.
    * 
    * @return  <code>true</code> if the operand opposite <code>ast</code> is
    *          nullable, else <code>false</code>.
    */
   private boolean isOppositeOperandNullable(HQLAst ast, int index)
   {
      boolean nullable = true;
      HQLAst target = (HQLAst) (index == 0
                                ? ast.getNextSibling()
                                : ast.getPrevSibling());
      HQLAst property = null;
      
      do
      {
         switch (target.getType())
         {
            case PROPERTY:
               property = target;
               break;
            case ALIAS:
               target = (HQLAst) target.getFirstChild();
               break;
            default:
               target = null;
         }
      } while (property == null && target != null);
      
      if (property != null)
      {
         nullable = !DatabaseManager.PRIMARY_KEY.equals(property.getText());
      }
      
      return nullable;
   }
   
   /**
    * Instances of this class are used as keys to cache instances of
    * <code>HQLPreprocessor</code>.
    */
   private static class CacheKey
   {
      /** Database instance */
      private final Database database;
      
      /** DMO implementation class, possibly <code>null</code> */
      private final Class<?> dmoClass;
      
      /** Where clause */
      private final String where;
      
      /** Query substitution parameter types */
      private final Type[] types;
      
      /** Whether query substitution parameters <em>may</em> be inlined */
      private final boolean inlinable;
      
      /**
       * Constructor.
       * 
       * @param   database
       *          Database instance.
       * @param   dmoClass
       *          DMO implementation class, possibly <code>null</code>.
       * @param   where
       *          Where clause.
       * @param   types
       *          Query substitution parameter types.
       * @param   inlinable
       *          Whether query substitution parameters <em>may</em> be
       *          inlined.
       */
      CacheKey(Database database,
               Class<?> dmoClass,
               String where,
               Type[] types,
               boolean inlinable)
      {
         this.database = database;
         this.dmoClass = dmoClass;
         this.where = where;
         this.types = types;
         this.inlinable = inlinable;
      }
      
      /**
       * Return a hash code consistent with {@link #equals}.
       * 
       * @return  Hash code.
       */
      public int hashCode()
      {
         int result = 17;
         
         result = 37 * result + database.hashCode();
         result = 37 * result + where.hashCode();
         if (dmoClass != null)
         {
            result = 37 * result + dmoClass.hashCode();
         }
         if (types != null)
         {
            for (Type t : types)
            {
               result = 37 * result + t.hashCode();
            }
         }
         result = 37 * result + (inlinable ? 0 : 1);

         return result;
      }
      
      /**
       * Check this object for equivalence with another instance of this
       * class.
       * 
       * @param   o
       *          Another instance of this class.
       * 
       * @return  <code>true</code> if instances are equivalent, else
       *          <code>false</code>.
       */
      public boolean equals(Object o)
      {
         CacheKey that = (CacheKey) o;
         
         if (inlinable != that.inlinable)
         {
            return false;
         }
         
         if (dmoClass != that.dmoClass)
         {
            return false;
         }
         
         if (!database.equals(that.database))
         {
            return false;
         }
         
         if (!where.equals(that.where))
         {
            return false;
         }
         
         if (types == null || that.types == null)
         {
            return (types == that.types);
         }
         
         int len1 = types.length;
         int len2 = that.types.length;
         
         if (len1 != len2)
         {
            return false;
         }
         
         for (int i = 0; i < len1; i++)
         {
            if (!types[i].equals(that.types[i]))
            {
               return false;
            }
         }
         
         return true;
      }
   }
}
